/*
 * Copyright 2008, Freescale Semiconductor, Inc
 * Andy Fleming
 *
 * Based vaguely on the Linux code
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <config.h>
#include <common.h>
#include <command.h>
#include <mmc.h>
#include <part.h>
#include <malloc.h>
#include <linux/list.h>
#include <div64.h>
#include <errno.h>

#include <drivers/mmc/rtkmmc.h>
#include <drivers/mmc/rtkmmc_generic.h>

/* Set block count limit because of 16 bit register limit on some hardware*/
#ifndef CONFIG_SYS_MMC_MAX_BLK_COUNT
#define CONFIG_SYS_MMC_MAX_BLK_COUNT 65535
#endif

#define RTK_HS400_FUNC

static struct list_head mmc_devices;
static int cur_dev_num = -1;

extern int mmc_data_sync( struct mmc * mmc );
void mmc_set_bus_timing(struct mmc *mmc, uint timing);

int __board_mmc_getcd(struct mmc *mmc) {
    return -1;
}

int board_mmc_getcd(struct mmc *mmc)__attribute__((weak,
    alias("__board_mmc_getcd")));

#ifdef CONFIG_MMC_BOUNCE_BUFFER
#ifdef CONFIG_RTK_MMC
#define RTK_MMC_DMA_MINALIGN 16
#endif

static int mmc_bounce_need_bounce(struct mmc_data *orig)
{
    ulong addr, len;

    if (orig->flags & MMC_DATA_READ)
        addr = (ulong)orig->dest;
    else
        addr = (ulong)orig->src;

#ifdef CONFIG_RTK_MMC
	if (addr % RTK_MMC_DMA_MINALIGN) {
        debug("MMC: Unaligned data destination address %08lx!\n", addr);
        return 1;
    }
#else
    if (addr % ARCH_DMA_MINALIGN) {
		debug("MMC: Unaligned data destination address %08lx!\n", addr);
        return 1;
    }
#endif

    len = (ulong)(orig->blocksize * orig->blocks);

#ifdef CONFIG_RTK_MMC
	if (len % RTK_MMC_DMA_MINALIGN) {
		debug("MMC: Unaligned data destination length %08lx!\n", len);
        return 1;
    }
#else
    if (len % ARCH_DMA_MINALIGN) {
		debug("MMC: Unaligned data destination length %08lx!\n", len);
        return 1;
    }
#endif

    return 0;
}

static int mmc_bounce_buffer_start(struct mmc_data *backup,
                    struct mmc_data *orig)
{
    ulong origlen, len;
    void *buffer;

    if (!orig)
        return 0;

    if (!mmc_bounce_need_bounce(orig))
        return 0;

    memcpy(backup, orig, sizeof(struct mmc_data));

    origlen = orig->blocksize * orig->blocks;

#ifdef CONFIG_RTK_MMC
	len = roundup(origlen, RTK_MMC_DMA_MINALIGN);
	buffer = memalign(RTK_MMC_DMA_MINALIGN, len);
#else
	len = roundup(origlen, ARCH_DMA_MINALIGN);
    buffer = memalign(ARCH_DMA_MINALIGN, len);
#endif
    if (!buffer) {
        puts("MMC: Error allocating MMC bounce buffer!\n");
        return 1;
    }

    if (orig->flags & MMC_DATA_READ) {
        orig->dest = buffer;
    } else {
        memcpy(buffer, orig->src, origlen);
        orig->src = buffer;
    }

    return 0;
}

static void mmc_bounce_buffer_stop(struct mmc_data *backup,
                    struct mmc_data *orig)
{
    ulong len;

    if (!orig)
        return;

    if (!mmc_bounce_need_bounce(backup))
        return;

    if (backup->flags & MMC_DATA_READ) {
        len = backup->blocksize * backup->blocks;
        memcpy(backup->dest, orig->dest, len);
        free(orig->dest);
        orig->dest = backup->dest;
    } else {
        free((void *)orig->src);
        orig->src = backup->src;
    }

    return;

}
#else
static inline int mmc_bounce_buffer_start(struct mmc_data *backup,
                    struct mmc_data *orig) { return 0; }
static inline void mmc_bounce_buffer_stop(struct mmc_data *backup,
                    struct mmc_data *orig) { }
#endif

int mmc_send_cmd(struct mmc *mmc, struct mmc_cmd *cmd, struct mmc_data *data)
{
    struct mmc_data backup;
    int ret;

    memset(&backup, 0, sizeof(backup));

    ret = mmc_bounce_buffer_start(&backup, data);
    if (ret)
        return ret;

#ifdef CONFIG_MMC_TRACE
    int i;
    u8 *ptr;

    printf("CMD_SEND:%d\n", cmd->cmdidx);
    printf("\t\tARG\t\t\t 0x%08X\n", cmd->cmdarg);
    printf("\t\tFLAG\t\t\t %d\n", cmd->flags);
    ret = mmc->send_cmd(mmc, cmd, data);
    switch (cmd->resp_type) {
        case MMC_RSP_NONE:
            printf("\t\tMMC_RSP_NONE\n");
            break;
        case MMC_RSP_R1:
            printf("\t\tMMC_RSP_R1,5,6,7 \t 0x%08X \n",
                cmd->response[0]);
            break;
        case MMC_RSP_R1b:
            printf("\t\tMMC_RSP_R1b\t\t 0x%08X \n",
                cmd->response[0]);
            break;
        case MMC_RSP_R2:
            printf("\t\tMMC_RSP_R2\t\t 0x%08X \n",
                cmd->response[0]);
            printf("\t\t          \t\t 0x%08X \n",
                cmd->response[1]);
            printf("\t\t          \t\t 0x%08X \n",
                cmd->response[2]);
            printf("\t\t          \t\t 0x%08X \n",
                cmd->response[3]);
            printf("\n");
            printf("\t\t\t\t\tDUMPING DATA\n");
            for (i = 0; i < 4; i++) {
                int j;
                printf("\t\t\t\t\t%03d - ", i*4);
                ptr = (u8 *)&cmd->response[i];
                ptr += 3;
                for (j = 0; j < 4; j++)
                    printf("%02X ", *ptr--);
                printf("\n");
            }
            break;
        case MMC_RSP_R3:
            printf("\t\tMMC_RSP_R3,4\t\t 0x%08X \n",
                cmd->response[0]);
            break;
        default:
            printf("\t\tERROR MMC rsp not supported\n");
            break;
    }
#else
    ret = mmc->send_cmd(mmc, cmd, data);
#endif
    mmc_bounce_buffer_stop(&backup, data);
    return ret;
}

int mmc_send_status(struct mmc *mmc, int timeout,int no_crc)
{
    struct mmc_cmd cmd;
    int err, retries = 5;
    unsigned int w_loop = 0;;
#ifdef CONFIG_MMC_TRACE
    int status;
#endif

    cmd.cmdidx = MMC_CMD_SEND_STATUS;
    cmd.resp_type = MMC_RSP_R1;
	if(no_crc)
		cmd.resp_type &= ~MMC_RSP_CRC;
    if (!mmc_host_is_spi(mmc))
        cmd.cmdarg = mmc->rca << 16;
    cmd.flags = 0;

    if(!timeout){
        w_loop = (500*10);
        printf("%s(%d)error func argument!!! set loop=%u\n",
                __func__,__LINE__,w_loop);
    }else{
        w_loop = (timeout*10);
    }
    do {
        err = mmc_send_cmd(mmc, &cmd, NULL);
        if (!err) {
            if ((cmd.response[0] & MMC_STATUS_RDY_FOR_DATA) &&
                (cmd.response[0] & MMC_STATUS_CURR_STATE) !=
                 MMC_STATE_PRG)
                break;
            else if (cmd.response[0] & MMC_STATUS_MASK) {
                printf("Status Error: 0x%08X\n",
                    cmd.response[0]);
                return COMM_ERR;
            }
        } else if (--retries < 0)
            return err;

        udelay(100);

    } while (w_loop--);

#ifdef CONFIG_MMC_TRACE
    status = (cmd.response[0] & MMC_STATUS_CURR_STATE) >> 9;
    printf("CURR STATE:%d\n", status);
#endif
    if (!w_loop) {
        printf("Timeout waiting card ready\n");
        return TIMEOUT;
    }

    return 0;
}

int mmc_set_blocklen(struct mmc *mmc, int len)
{
#if 1
    return 0;
#else
    struct mmc_cmd cmd;

    cmd.cmdidx = MMC_CMD_SET_BLOCKLEN;
    cmd.resp_type = MMC_RSP_R1;
    cmd.cmdarg = len;
    cmd.flags = 0;

    return mmc_send_cmd(mmc, &cmd, NULL);
#endif
}

struct mmc *find_mmc_device(int dev_num)
{
    struct mmc *m;
    struct list_head *entry;

    list_for_each(entry, &mmc_devices) {
        m = list_entry(entry, struct mmc, link);

        if (m->block_dev.dev == dev_num)
            return m;
    }

    printf("MMC Device %d not found\n", dev_num);

    return NULL;
}

#define TRIM_ERASE
static ulong mmc_erase_t(struct mmc *mmc, ulong start, lbaint_t blkcnt)
{
    struct mmc_cmd cmd;
    ulong end;
    int err, start_cmd, end_cmd;

    if (mmc->high_capacity){
        end = start + blkcnt - 1;
    }else {
        end = (start + blkcnt - 1) * mmc->write_bl_len;
        start *= mmc->write_bl_len;
    }

    if (IS_SD(mmc)) {
        start_cmd = SD_CMD_ERASE_WR_BLK_START;
        end_cmd = SD_CMD_ERASE_WR_BLK_END;
    } else {
        start_cmd = MMC_CMD_ERASE_GROUP_START;
        end_cmd = MMC_CMD_ERASE_GROUP_END;
    }
/*
    printf("erasing... (start=0x%08x end=0x%08x count=%u)\n",
		(unsigned int)start, (unsigned int)end, (unsigned int)blkcnt);
*/
    cmd.cmdidx = start_cmd;
    cmd.cmdarg = start;
    cmd.resp_type = MMC_RSP_R1;
    cmd.flags = 0;

    err = mmc_send_cmd(mmc, &cmd, NULL);
    if (err)
        goto err_out;

    cmd.cmdidx = end_cmd;
    cmd.cmdarg = end;

    err = mmc_send_cmd(mmc, &cmd, NULL);
    if (err)
        goto err_out;

    cmd.cmdidx = MMC_CMD_ERASE;

	if(mmc->sec_support & (0x01UL<<4)){
    cmd.cmdarg = TRIM_OPER;
	}else{
    cmd.cmdarg = SECURE_ERASE;
	}
    cmd.resp_type = MMC_RSP_R1b;

    err = mmc_send_cmd(mmc, &cmd, NULL);
    if (err)
        goto err_out;

    return 0;

err_out:
    puts("mmc erase failed\n");
    return err;
}

static unsigned long
mmc_berase(int dev_num, unsigned long start, lbaint_t blkcnt)
{
	int err = 0;
	struct mmc *mmc = find_mmc_device(dev_num);
	lbaint_t blk = 0, blk_r = 0;
	int timeout = 1000;
	unsigned int loop_show = 0;

	if (!mmc)
		return -1;

	printf("\n\nSEC_FEATURE_SUPPORT=0x%08x.\n",mmc->sec_support);
	if(mmc->sec_support & (0x01UL<<4)){
		printf("TRIM mode\n");
		if(start+blkcnt > (mmc->capacity>>9)){
			blkcnt =  (mmc->capacity>>9)-start;
			printf("\n\nCaution! Your erase size beyond capacity.\n"
			"         recommend to modify your command to:\n"
			"         mmc erase %lx %lx\n",start,blkcnt);
		}
	}else{
		printf("ERASE GROUP mode\n");
		if ((start % mmc->erase_grp_size) || (blkcnt % mmc->erase_grp_size))
			printf("\n\nCaution! Your devices Erase group is 0x%x\n"
				"The erase range would be change to 0x%lx~0x%lx\n\n",
				mmc->erase_grp_size, start & ~(mmc->erase_grp_size - 1),
				((start + blkcnt + mmc->erase_grp_size)
				& ~(mmc->erase_grp_size - 1)) - 1);
	}

#define MAX_LOOP_SHOW 0x100
	loop_show = 0;
	printf("\nerasing...... (start=0x%08x ",(unsigned int)start);

	while (blk < blkcnt) {
#if 0
		if (mmc->erase_grp_size) {
			if ((((blkcnt - blk))/(mmc->erase_grp_size)) >= 1000) {
				blk_r = (mmc->erase_grp_size) * 1000;
			} else if ((((blkcnt - blk))/(mmc->erase_grp_size)) >= 100) {
				blk_r = (mmc->erase_grp_size) * 100;
			} else if ((((blkcnt - blk))/(mmc->erase_grp_size)) >= 10) {
				blk_r = (mmc->erase_grp_size) * 10;
			} else {
				blk_r = ((blkcnt - blk) > mmc->erase_grp_size) ?
				mmc->erase_grp_size : (blkcnt - blk);
			}
		} else {
			blk_r = ((blkcnt - blk) > mmc->erase_grp_size) ?
			mmc->erase_grp_size : (blkcnt - blk);
		}
#else
		blk_r = ((blkcnt - blk) > mmc->erase_grp_size) ?
		mmc->erase_grp_size : (blkcnt - blk);
#endif

		if(loop_show >= (MAX_LOOP_SHOW-1)){
			loop_show = 0;
		}else{
			loop_show++;
			if(loop_show & 0x10)
			printf(".");
		}

		err = mmc_erase_t(mmc, start + blk, blk_r);
		if (err)
			break;

		blk += blk_r;

		/* Waiting for the ready status */
		timeout = 10000;
		if (mmc_send_status(mmc, timeout,0))
			return 0;
	}
	printf("\n");
	return blk;
}

static ulong
mmc_write_blocks(struct mmc *mmc, ulong start, lbaint_t blkcnt, const void*src)
{
    struct mmc_cmd cmd;
    struct mmc_data data;
    int retry = 5;
    int timeout = 2000;

    if ((start + blkcnt) > mmc->block_dev.lba) {
        printf("MMC: block number 0x%lx exceeds max(0x%lx)\n",
            start + blkcnt, mmc->block_dev.lba);
        return 0;
    }

#ifndef CONFIG_QC_VERIFY
    RETRY:
#endif

    if (blkcnt > 1)
        cmd.cmdidx = MMC_CMD_WRITE_MULTIPLE_BLOCK;
    else
        cmd.cmdidx = MMC_CMD_WRITE_SINGLE_BLOCK;

    if (mmc->high_capacity)
        cmd.cmdarg = start;
    else
        cmd.cmdarg = start * mmc->write_bl_len;

    cmd.resp_type = MMC_RSP_R1;
    cmd.flags = 0;

    data.src = src;
    data.blocks = blkcnt;
    data.blocksize = mmc->write_bl_len;
    data.flags = MMC_DATA_WRITE;

    if (mmc_send_cmd(mmc, &cmd, &data)) {
		/*
		Add re-initial flow here~~
		*/
		if(retry){
			printf("%s(%d)Call Re-Initial!!(retry%d)\n",__func__,__LINE__,retry);
			retry--;
#ifndef CONFIG_QC_VERIFY
			if(mmc->host_caps & MMC_MODE_HS_400MHz){
				printf("%s(%d)disable HS400 mode!!!(retry%d)\n",__func__,__LINE__,retry);
				mmc->host_caps &= ~MMC_MODE_HS_400MHz;
				export_rtk_emmc_chk_max_xfer_blk(mmc,1);
			}else{

				if(!retry
				 && mmc->host_caps & MMC_MODE_HS_200MHz)
				{
					printf("%s(%d)disable HS200 mode!!!(retry%d)\n",__func__,__LINE__,retry);
					mmc->host_caps &= ~MMC_MODE_HS_200MHz;
					export_rtk_emmc_chk_max_xfer_blk(mmc,1);
					mmc->f_max = 50000000;
					retry = 5;
				}
			}
			mmc->has_init = 0;
			mmc->timing = 0;
			mmc->bus_width = 0;
			mmc_init(mmc);
			goto RETRY;
#endif
		}else{
			printf("%s(%d)mmc write retry fail! give up!\n",__func__,__LINE__);
            return 0;
        }
	}

#ifndef AUTO_SEND_STOP
	/* RTK driver have auto stop func */

    /* SPI multiblock writes terminate using a special
     * token, not a STOP_TRANSMISSION request.
     */
    if (!mmc_host_is_spi(mmc) && blkcnt > 1) {
        cmd.cmdidx = MMC_CMD_STOP_TRANSMISSION;
        cmd.cmdarg = 0;
        cmd.resp_type = MMC_RSP_R1b;
        cmd.flags = 0;
        if (mmc_send_cmd(mmc, &cmd, NULL)) {
            printf("mmc fail to send stop cmd\n");
            return 0;
        }
    }
#endif
    /* Waiting for the ready status */
	if (mmc_send_status(mmc, timeout,0))
        return 0;

    return blkcnt;
}

static ulong
mmc_bwrite(int dev_num, ulong start, lbaint_t blkcnt, const void*src)
{
    lbaint_t cur, blocks_todo = blkcnt;

    struct mmc *mmc = find_mmc_device(dev_num);
    if (!mmc)
        return 0;

    if (mmc_set_blocklen(mmc, mmc->write_bl_len))
        return 0;

    do {
        cur = (blocks_todo > mmc->b_max) ?  mmc->b_max : blocks_todo;
        if(mmc_write_blocks(mmc, start, cur, src) != cur)
            return 0;
        blocks_todo -= cur;
        start += cur;
        src += cur * mmc->write_bl_len;
    } while (blocks_todo > 0);

    return blkcnt;
}

int mmc_read_blocks(struct mmc *mmc, void *dst, ulong start, lbaint_t blkcnt)
{
	struct mmc_cmd cmd;
	struct mmc_data data;
#ifdef CONFIG_QC_VERIFY
    int retry = 0;
#else
    int retry = 5;
#endif
RETRY:
	if (blkcnt > 1)
		cmd.cmdidx = MMC_CMD_READ_MULTIPLE_BLOCK;
	else
		cmd.cmdidx = MMC_CMD_READ_SINGLE_BLOCK;

	if (mmc->high_capacity)
		cmd.cmdarg = start;
	else
		cmd.cmdarg = start * mmc->read_bl_len;

	cmd.resp_type = MMC_RSP_R1;
	cmd.flags = 0;

	data.dest = dst;
	data.blocks = blkcnt;
	data.blocksize = mmc->read_bl_len;
	data.flags = MMC_DATA_READ;

	if (mmc_send_cmd(mmc, &cmd, &data)){
		/*
		Add re-initial flow here~~
		*/
		if(retry){
			printf("%s(%d)Call Re-Initial!!(retry%d)\n",__func__,__LINE__,retry);
			if(mmc->host_caps & MMC_MODE_HS_400MHz){
				printf("%s(%d)disable HS400 mode!!!(retry%d)\n",__func__,__LINE__,retry);
				mmc->host_caps &= ~MMC_MODE_HS_400MHz;
				mmc->b_max = EMMC_MAX_XFER_BLK>>1;
				SET_SPEED_MODE(DIS_HS400);
			}else{
				if( (retry == 1)
				 && (mmc->host_caps & MMC_MODE_HS_200MHz) )
				{
					printf("%s(%d)disable HS200 mode!!!(retry%d)\n",__func__,__LINE__,retry);
					mmc->host_caps &= ~MMC_MODE_HS_200MHz;
					mmc->b_max = 600;
					mmc->f_max = 50000000;
					SET_SPEED_MODE(DIS_HS200);
					retry = 5;
				}
			}
			retry--;
			mmc->has_init = 0;
			mmc->timing = 0;
			mmc->bus_width = 0;
			mmc_init(mmc);
			goto RETRY;

		}else{
			printf("%s(%d)mmc read retry fail! give up!\n",__func__,__LINE__);
			return 0;
		}
	}

#ifndef AUTO_SEND_STOP
	/* RTK driver have auto stop func */
	if (blkcnt > 1) {
		cmd.cmdidx = MMC_CMD_STOP_TRANSMISSION;
		cmd.cmdarg = 0;
		cmd.resp_type = MMC_RSP_R1b;
		cmd.flags = 0;
		if (mmc_send_cmd(mmc, &cmd, NULL)) {
			printf("mmc fail to send stop cmd\n");
			return 0;
		}
	}
#endif
	return blkcnt;
}

static ulong mmc_bread(int dev_num, ulong start, lbaint_t blkcnt, void *dst)
{
	lbaint_t cur, blocks_todo = blkcnt;
	struct mmc *mmc = find_mmc_device(dev_num);

	if (!mmc)
		return 0;

	if (blkcnt == 0)
		return 0;

	if ((start + blkcnt) > mmc->block_dev.lba) {
		printf("MMC: block number 0x%lx exceeds max(0x%lx)\n",
			start + blkcnt, mmc->block_dev.lba);
		return 0;
	}

	if (mmc_set_blocklen(mmc, mmc->read_bl_len))
		return 0;

	do {
		cur = (blocks_todo > mmc->b_max) ?  mmc->b_max : blocks_todo;
		if(mmc_read_blocks(mmc, dst, start, cur) != cur)
			return 0;
		blocks_todo -= cur;
		start += cur;
		dst += cur * mmc->read_bl_len;
	} while (blocks_todo > 0);

	return blkcnt;
}

#if defined(CONFIG_LG_CHG)
static char *_pu1BlkBuf = NULL;

static unsigned long mmc_byteread(int dev_num, u64 start, unsigned long bytecnt, void *dst)
{
	u64 u4OffsetInBlk;
	u64 u4BlkOffset;
	unsigned long u4RestSize = bytecnt;
	unsigned long u4MemPtr = (ulong)dst;
	struct mmc *mmc = find_mmc_device(dev_num);

	if (!mmc)
		return 0;

	if (!_pu1BlkBuf){
		_pu1BlkBuf = (char *)malloc(mmc->read_bl_len + 0x20);
		_pu1BlkBuf = (char *)((((uint)(_pu1BlkBuf))&(~0x1F)) + 0x20);
	}

	u4OffsetInBlk = start & (mmc->read_bl_len - 1);
	u4BlkOffset   = start - u4OffsetInBlk;

	// Handle offset non-block-size-aligned case
	if (u4OffsetInBlk){
		ulong u4OpSize;

		if (mmc_bread(dev_num, u4BlkOffset/mmc->read_bl_len, 1, _pu1BlkBuf) != 1)
			return 0;

		if (u4RestSize >= (mmc->read_bl_len - u4OffsetInBlk))
			u4OpSize = mmc->read_bl_len - u4OffsetInBlk;
		else
			u4OpSize = u4RestSize;

		memcpy((void *)u4MemPtr, (void *)((char *)_pu1BlkBuf + u4OffsetInBlk), u4OpSize);

		u4RestSize  -= u4OpSize;
		u4MemPtr    += u4OpSize;
		u4BlkOffset += mmc->read_bl_len;
	}

	// Handle offset & size block-size-aligned case
	if (u4RestSize >= mmc->read_bl_len){
		ulong u4BlkCnt = u4RestSize/mmc->read_bl_len;

		if (mmc_bread(dev_num, u4BlkOffset/mmc->read_bl_len, u4BlkCnt, (void *)u4MemPtr) != u4BlkCnt)
			return 0;

		u4RestSize  -= (u4BlkCnt * mmc->read_bl_len);
		u4MemPtr    += (u4BlkCnt * mmc->read_bl_len);
		u4BlkOffset += (u4BlkCnt * mmc->read_bl_len);

	}

	// Handle size none-block-size-aligned case
	if (u4RestSize){
		if (mmc_bread(dev_num, u4BlkOffset/mmc->read_bl_len, 1, _pu1BlkBuf) != 1)
			return 0;

		memcpy((void *)u4MemPtr, (void *)_pu1BlkBuf, u4RestSize);
	}

	return bytecnt;
}

static unsigned long mmc_bytewrite(int dev_num, u64 start, unsigned long bytecnt, const void *dst)
{
	u64 u4OffsetInBlk, u4BlkOffset;
	unsigned long u4RestSize = bytecnt;
	unsigned long u4MemPtr = (ulong)dst;

	struct mmc *mmc = find_mmc_device(dev_num);
	if (!mmc)
		return 0;

	if (!_pu1BlkBuf){
		_pu1BlkBuf = (char *)malloc(mmc->write_bl_len + 0x20);
		_pu1BlkBuf = (char *)((((uint)(_pu1BlkBuf))&(~0x1F)) + 0x20);
	}

	u4OffsetInBlk = start & (mmc->write_bl_len - 1);
	u4BlkOffset = start - u4OffsetInBlk;

	// Handle offset non-block-size-aligned case
	if (u4OffsetInBlk){
		ulong u4OpSize;

		if (mmc_bread(dev_num, u4BlkOffset/mmc->write_bl_len, 1, _pu1BlkBuf) != 1)
			return 0;

		if (u4RestSize >= (mmc->write_bl_len - u4OffsetInBlk))
			u4OpSize = mmc->write_bl_len - u4OffsetInBlk;
		else
			u4OpSize = u4RestSize;

		memcpy((void *)((char *)_pu1BlkBuf + u4OffsetInBlk), (void *)u4MemPtr, u4OpSize);

		if (mmc_bwrite(dev_num, u4BlkOffset/mmc->write_bl_len, 1, _pu1BlkBuf) != 1)
			return 0;

		u4RestSize  -= u4OpSize;
		u4MemPtr    += u4OpSize;
		u4BlkOffset += mmc->write_bl_len;
	}

	// Handle offset & size block-size-aligned case
	if (u4RestSize >= mmc->read_bl_len){
		ulong u4BlkCnt = u4RestSize/mmc->write_bl_len;

		if (mmc_bwrite(dev_num, u4BlkOffset/mmc->write_bl_len, u4BlkCnt, (void *)u4MemPtr) != u4BlkCnt)
			return 0;

		u4RestSize  -= (u4BlkCnt * mmc->write_bl_len);
		u4MemPtr    += (u4BlkCnt * mmc->write_bl_len);
		u4BlkOffset += (u4BlkCnt * mmc->write_bl_len);
	}

	// Handle size none-block-size-aligned case
	if (u4RestSize){
		if (mmc_bread(dev_num, u4BlkOffset/mmc->write_bl_len, 1, _pu1BlkBuf) != 1)
			return 0;

		memcpy((void *)_pu1BlkBuf, (void *)u4MemPtr, u4RestSize);

		if (mmc_bwrite(dev_num, u4BlkOffset/mmc->write_bl_len, 1, _pu1BlkBuf) != 1)
			return 0;

	}

	return bytecnt;
}

static unsigned long
mmc_erase_type(struct mmc *mmc, unsigned long start, lbaint_t blkcnt,unsigned int type)
{
	struct mmc_cmd cmd;
	ulong end;
	int err, start_cmd, end_cmd;

	if (mmc->high_capacity)
		end = start + blkcnt - 1;
	else {
		end = (start + blkcnt - 1) * mmc->write_bl_len;
		start *= mmc->write_bl_len;
	}

	if (IS_SD(mmc)) {
		start_cmd = SD_CMD_ERASE_WR_BLK_START;
		end_cmd = SD_CMD_ERASE_WR_BLK_END;
	} else {
		start_cmd = MMC_CMD_ERASE_GROUP_START;
		end_cmd = MMC_CMD_ERASE_GROUP_END;
	}

	cmd.cmdidx = start_cmd;
	cmd.cmdarg = start;
	cmd.resp_type = MMC_RSP_R1;
	cmd.flags = 0;

	err = mmc_send_cmd(mmc, &cmd, NULL);
	if (err)
		goto err_out;

	cmd.cmdidx = end_cmd;
	cmd.cmdarg = end;

	err = mmc_send_cmd(mmc, &cmd, NULL);
	if (err)
		goto err_out;

	cmd.cmdidx = MMC_CMD_ERASE;
	cmd.cmdarg = type;
	cmd.resp_type = MMC_RSP_R1b;

	err = mmc_send_cmd(mmc, &cmd, NULL);
	if (err)
		goto err_out;

	return 0;

err_out:
	puts("mmc erase failed\n");
	return err;
}

static unsigned long
mmc_berase_type(int dev_num, unsigned long start, lbaint_t blkcnt,unsigned int type)
{
	struct mmc *mmc = find_mmc_device(dev_num);
	int err = 0;
	lbaint_t blk = 0, blk_r = 0;

	if (!mmc)
		return -1;

	if(!(type & TRIM_OPER)){
		if ((start % mmc->erase_grp_size) || (blkcnt % mmc->erase_grp_size))
			printf("\n\nCaution! Your devices Erase group is 0x%x\n"
				"The erase range would be change to 0x%lx~0x%lx\n\n",
				mmc->erase_grp_size, start & ~(mmc->erase_grp_size - 1),
				((start + blkcnt + mmc->erase_grp_size)
				& ~(mmc->erase_grp_size - 1)) - 1);
	}

	while (blk < blkcnt) {

		blk_r = ((blkcnt - blk) > mmc->erase_grp_size) ?
			mmc->erase_grp_size : (blkcnt - blk);
		err = mmc_erase_type(mmc, start + blk, blk_r,type);
		if (err)
			break;

		blk += blk_r;
	}

	return blk;
}
#endif	// #if defined(CONFIG_LG_CHG)

int mmc_go_idle(struct mmc* mmc)
{
    struct mmc_cmd cmd;
    int err;

    udelay(1000);

    cmd.cmdidx = MMC_CMD_GO_IDLE_STATE;
    cmd.cmdarg = 0;
    cmd.resp_type = MMC_RSP_NONE;
    cmd.flags = 0;

    err = mmc_send_cmd(mmc, &cmd, NULL);

    if (err)
        return err;

    udelay(2000);

    return 0;
}

int
sd_send_op_cond(struct mmc *mmc)
{
    int timeout = 1000;
    int err;
    struct mmc_cmd cmd;

    do {
        cmd.cmdidx = MMC_CMD_APP_CMD;
        cmd.resp_type = MMC_RSP_R1;
        cmd.cmdarg = 0;
        cmd.flags = 0;

        err = mmc_send_cmd(mmc, &cmd, NULL);

        if (err)
            return err;

        cmd.cmdidx = SD_CMD_APP_SEND_OP_COND;
        cmd.resp_type = MMC_RSP_R3;

        /*
         * Most cards do not answer if some reserved bits
         * in the ocr are set. However, Some controller
         * can set bit 7 (reserved for low voltages), but
         * how to manage low voltages SD card is not yet
         * specified.
         */
        cmd.cmdarg = mmc_host_is_spi(mmc) ? 0 :
            (mmc->voltages & 0xff8000);

        if (mmc->version == SD_VERSION_2)
            cmd.cmdarg |= OCR_HCS;

        err = mmc_send_cmd(mmc, &cmd, NULL);

        if (err)
            return err;

        udelay(1000);
    } while ((!(cmd.response[0] & OCR_BUSY)) && timeout--);

    if (timeout <= 0)
        return UNUSABLE_ERR;

    if (mmc->version != SD_VERSION_2)
        mmc->version = SD_VERSION_1_0;

    if (mmc_host_is_spi(mmc)) { /* read OCR for spi */
        cmd.cmdidx = MMC_CMD_SPI_READ_OCR;
        cmd.resp_type = MMC_RSP_R3;
        cmd.cmdarg = 0;
        cmd.flags = 0;

        err = mmc_send_cmd(mmc, &cmd, NULL);

        if (err)
            return err;
    }

    mmc->ocr = cmd.response[0];

    mmc->high_capacity = ((mmc->ocr & OCR_HCS) == OCR_HCS);
    mmc->rca = 0;

    return 0;
}

int mmc_send_op_cond(struct mmc *mmc)
{
    int timeout = 10000;
    struct mmc_cmd cmd;
    int err;

    /* Some cards seem to need this */
    mmc_go_idle(mmc);

    /* Asking to the card its capabilities */
    cmd.cmdidx = MMC_CMD_SEND_OP_COND;
    cmd.resp_type = MMC_RSP_R3;
    cmd.cmdarg = 0;
    cmd.flags = 0;

    err = mmc_send_cmd(mmc, &cmd, NULL);

    if (err)
        return err;

    udelay(1000);

    do {
        cmd.cmdidx = MMC_CMD_SEND_OP_COND;
        cmd.resp_type = MMC_RSP_R3;
        cmd.cmdarg = (mmc_host_is_spi(mmc) ? 0 :
                (mmc->voltages &
                (cmd.response[0] & OCR_VOLTAGE_MASK)) |
                (cmd.response[0] & OCR_ACCESS_MODE));

        if (mmc->host_caps & MMC_MODE_HC)
            cmd.cmdarg |= OCR_HCS;

        cmd.flags = 0;

        err = mmc_send_cmd(mmc, &cmd, NULL);

        if (err)
            return err;

        udelay(1000);
    } while (!(cmd.response[0] & OCR_BUSY) && timeout--);

    if (timeout <= 0)
        return UNUSABLE_ERR;

    if (mmc_host_is_spi(mmc)) { /* read OCR for spi */
        cmd.cmdidx = MMC_CMD_SPI_READ_OCR;
        cmd.resp_type = MMC_RSP_R3;
        cmd.cmdarg = 0;
        cmd.flags = 0;

        err = mmc_send_cmd(mmc, &cmd, NULL);

        if (err)
            return err;
    }

    mmc->version = MMC_VERSION_UNKNOWN;
    mmc->ocr = cmd.response[0];

    mmc->high_capacity = ((mmc->ocr & OCR_HCS) == OCR_HCS);
    mmc->rca = 0;

    return 0;
}


int mmc_send_ext_csd(struct mmc *mmc, char *ext_csd)
{
    struct mmc_cmd cmd;
    struct mmc_data data;
    int err;

    /* Get the Card Status Register */
    cmd.cmdidx = MMC_CMD_SEND_EXT_CSD;
    cmd.resp_type = MMC_RSP_R1;
    cmd.cmdarg = 0;
    cmd.flags = 0;

    data.dest = ext_csd;
    data.blocks = 1;
    data.blocksize = 512;
    data.flags = MMC_DATA_READ;

    err = mmc_send_cmd(mmc, &cmd, &data);

    return err;
}

int export_mmc_send_ext_csd(char *ext_csd)
{
	int curr_device = 0;
	struct mmc *mmc = NULL;

	mmc = find_mmc_device(curr_device);

	return  mmc_send_ext_csd(mmc, ext_csd);
}

static char *malloc_aligned(unsigned int size,int alignment)
{
	char *ptr = malloc(size + alignment-1);

	if (!ptr)
		return NULL;

	if((unsigned int)ptr % alignment)
		ptr = (char *)((unsigned int)ptr - (unsigned int)((unsigned int)ptr % alignment) + alignment);

	return ptr;
}

//#define __DEBUG__

static int toshiba_smart_report(struct mmc *mmc)
{
	struct mmc_cmd cmd;
	struct mmc_data data;
	int err = 0;
#ifdef __DEBUG__
	int i=0;
#endif

	char *buf = NULL;

	if(((mmc->cid[0] >> 24) & 0xff) != 0x11){
		printf("This is toshiba emmc smart report function![mid = 0x%x]\n",((mmc->cid[0] >> 24) & 0xff));
		return -1;
	}

	buf = malloc_aligned(512, ARCH_DMA_MINALIGN);

	if ((unsigned int)buf % ARCH_DMA_MINALIGN) {
		printf("MMC: Unaligned data destination address %08lx!\n", (unsigned long int)buf);
		return -1;
	}

	memset(buf, 0 , 512);
	/* TOSHIBA 19nm ver 4 41alpha Erase Count Checking Sequence.pdf */
	/*
    	TOSHIBA 19nm ver 4 41alpha Erase Count Checking Sequence.pdf
    	Device Health Command Format
    	0x000         1    Sub Command Number (0x07)
    	0x001 V 0x003 3    Dummy Data (0x00)
    	0x004 V 0x007 4    Defined Password for Device Health 19nm v4.41\ (0x6E-0x2D-0x66-0xE0)
    	0x009 V 0x1FF 504  Reserved
	*/
	buf[0] = 0x07;
	buf[4] = 0x6E;
	buf[5] = 0x2D;
	buf[6] = 0x66;
	buf[7] = 0xE0;

	//init cmd
	cmd.cmdidx = 56;
	cmd.resp_type = MMC_RSP_R1;
	cmd.cmdarg = 0;
	cmd.flags = 0;

	data.src = buf;
	data.blocks = 1;
	data.blocksize = 512;
	data.flags = MMC_DATA_WRITE;

	err = mmc_send_cmd(mmc, &cmd, &data);
	if(err){
		printf("unable send TSB vendor life report cmd(ret_err=%d)\n", err);
		return err;
    	}

	udelay(500);

	/* get TSB vendor life report */
	memset(buf, 0x5a , 512);

	//init cmd
	cmd.cmdidx = 56;
	cmd.resp_type = MMC_RSP_R1;
	cmd.cmdarg = 1;
	cmd.flags = 0;

	data.dest = buf;
	data.blocks = 1;
	data.blocksize = 512;
	data.flags = MMC_DATA_READ;

	err = mmc_send_cmd(mmc, &cmd, &data);
	if(err){
		printf("get TSB vendor life report cmd error(ret_err=%d)\n", err);
		return err;
	}

#ifdef __DEBUG__
	//show info
	for(i=0;i< (512/8) ;i++) {
		printf("[%03x~%03x] = 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
			(i*8),((i*8)+7),
			buf[(i*8)+0], buf[(i*8)+1], buf[(i*8)+2], buf[(i*8)+3],
			buf[(i*8)+4], buf[(i*8)+5], buf[(i*8)+6], buf[(i*8)+7]);
	}
#endif

	//sanity check
	if(buf[0] != 0x07){
		printf("Sub command no error![cmd no = 0x%x]!\n", buf[0]);
		return -1;
	}

	if(buf[3] != 0xA0){
		printf("Command Status error![cmd status = 0x%x]!\n", buf[3]);
		return -1;
	}

#define BYTE(d,n)		(((d) >> ((n) << 3)) & 0xFF)
#define SWAPEND32(d)	((BYTE(d,0) << 24) | (BYTE(d,1) << 16) | (BYTE(d,2) << 8) | (BYTE(d,3) << 0))

	printf("Max Erase Count in MLC Area = 0x%x!\n", SWAPEND32(*((unsigned int *)((unsigned int)buf+4))));
	printf("Average Erase Count in MLC Area = 0x%x!\n", SWAPEND32(*((unsigned int *)((unsigned int)buf+8))));
	printf("Max Erase Count in pSLC Area = 0x%x!\n", SWAPEND32(*((unsigned int *)((unsigned int)buf+0xc))));
	printf("Average Erase Count in pSLC Area = 0x%x!\n", SWAPEND32(*((unsigned int *)((unsigned int)buf+0x10))));

	return 0;
}

static int micron_smart_report(struct mmc *mmc)
{
	struct mmc_cmd cmd;
	struct mmc_data data;
	unsigned int arg;
#ifdef __DEBUG__
	int i = 0;
#endif
	char *buf = NULL;
	int err =0;

	if(((mmc->cid[0] >> 24) & 0xff) != 0xfe){
		printf("This is Micron emmc smart report function![mid = 0x%x]\n",((mmc->cid[0] >> 24) & 0xff));
		return -1;
	}

	buf = malloc_aligned(512, ARCH_DMA_MINALIGN);

	if ((unsigned int)buf % ARCH_DMA_MINALIGN) {
		printf("MMC: Unaligned data destination address %08lx!\n", (unsigned long int)buf);
		return -1;
	}

	memset(buf, 0 , 512);

	/* depend on tnfc19_emmc_health_status_command.pdf
	 * chapter Block Erase Count Retrieve
	 *     Index = 10h: MIN/MAX/AVG erase count
	 *     Index = 11h: MIN/MAX/AVG erase count in the MLC area
	 *     Index = 12h: MIN/MAX/AVG erase count in the SLC area
	 */
	arg = (0x11U<1) | 0x01;	// first read MLC area

	//init cmd
	cmd.cmdidx = 56;
	cmd.resp_type = MMC_RSP_R1;
	cmd.cmdarg = arg;
	cmd.flags = 0;

	data.dest = buf;
	data.blocks = 1;
	data.blocksize = 512;
	data.flags = MMC_DATA_READ;

	err = mmc_send_cmd(mmc, &cmd, &data);
	if(err){
		printf("exe emmc health status cmd error(ret_err=%d)\n", err);
		return err;
	}else{
#ifdef __DEBUG__
	//show info
	for(i=0;i< (512/8) ;i++) {
		printf("[%03x~%03x] = 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
			(i*8),((i*8)+7),
			buf[(i*8)+0], buf[(i*8)+1], buf[(i*8)+2], buf[(i*8)+3],
			buf[(i*8)+4], buf[(i*8)+5], buf[(i*8)+6], buf[(i*8)+7]);
	}
#endif
		// info show
		printf("MIN erase count in MLC area = 0x%x !\n", buf[0]<<8 | buf[1]);
		printf("MAX erase count in MLC area = 0x%x !\n", buf[2]<<8 | buf[3]);
		printf("AVG erase count in MLC area = 0x%x !\n", buf[4]<<8 | buf[5]);
	}

	memset(buf, 0 , 512);

	arg = (0x12U<1) | 0x01;	// second read SLC area

	//init cmd
	cmd.cmdidx = 56;
	cmd.resp_type = MMC_RSP_R1;
	cmd.cmdarg = arg;
	cmd.flags = 0;

	data.dest = buf;
	data.blocks = 1;
	data.blocksize = 512;
	data.flags = MMC_DATA_READ;

	err = mmc_send_cmd(mmc, &cmd, &data);
	if(err){
		printf("exe emmc health status cmd error(ret_err=%d)", err);
		return err;
	}else{
#ifdef __DEBUG__
		//show info
		for(i=0;i< (512/8) ;i++) {
			printf("[%03x~%03x] = 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
				(i*8),((i*8)+7),
				buf[(i*8)+0], buf[(i*8)+1], buf[(i*8)+2], buf[(i*8)+3],
				buf[(i*8)+4], buf[(i*8)+5], buf[(i*8)+6], buf[(i*8)+7]);
		}
#endif
		// info show
		printf("MIN erase count in SLC area = 0x%x !\n", buf[0]<<8 | buf[1]);
		printf("MAX erase count in SLC area = 0x%x !\n", buf[2]<<8 | buf[3]);
		printf("AVG erase count in SLC area = 0x%x !\n", buf[4]<<8 | buf[5]);
	}

	return 0;
}

static int hynix_smart_report(struct mmc *mmc)
{
	struct mmc_cmd cmd;
	struct mmc_data data;
	unsigned int arg;
#ifdef __DEBUG__
	int i = 0;
#endif
	char *buf = NULL;
	int err =0;

	if(((mmc->cid[0] >> 24) & 0xff) != 0x90){
		printf("This is Hynix emmc smart report function![mid = 0x%x]\n",((mmc->cid[0] >> 24) & 0xff));
		return -1;
	}

	buf = malloc_aligned(512, ARCH_DMA_MINALIGN);

	if ((unsigned int)buf % ARCH_DMA_MINALIGN) {
		printf("MMC: Unaligned data destination address %08lx!\n", (unsigned long int)buf);
		return -1;
	}

	memset(buf, 0 , 512);

	//step1
	arg = 0x534d4900;
	cmd.cmdidx = 60;
	cmd.resp_type = MMC_RSP_R1;
	cmd.cmdarg = arg;
	cmd.flags = 0;

	err = mmc_send_cmd(mmc, &cmd, NULL);
	if(err){
		printf("step1:exe emmc health status cmd60 error(ret_err=%d)", err);
		return err;
	}

	//step2
	arg = 0x48525054;
	cmd.cmdidx = 60;
	cmd.resp_type = MMC_RSP_R1;
	cmd.cmdarg = arg;
	cmd.flags = 0;

	err = mmc_send_cmd(mmc, &cmd, NULL);
	if(err){
		printf("step2:exe emmc health status cmd60 error(ret_err=%d)", err);
		return err;
	}

	//step3
	arg = 0x0;
	cmd.cmdidx = 8;
	cmd.resp_type = MMC_RSP_R1;
	cmd.cmdarg = arg;
	cmd.flags = 0;

	data.dest = buf;
	data.blocks = 1;
	data.blocksize = 512;
	data.flags = MMC_DATA_READ;

	err = mmc_send_cmd(mmc, &cmd, &data);
	if(err){
		printf("step3:exe emmc health status cmd8 error(ret_err=%d)", err);
		return err;
	}

#ifdef __DEBUG__
	//show info
	for(i=0;i< (512/8) ;i++) {
		printf("[%03x~%03x] = 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
			(i*8),((i*8)+7),
			buf[(i*8)+0], buf[(i*8)+1], buf[(i*8)+2], buf[(i*8)+3],
			buf[(i*8)+4], buf[(i*8)+5], buf[(i*8)+6], buf[(i*8)+7]);
	}
#endif

	printf("Number of reserved blocks for MLC area = 0x%x !\n", *((unsigned int *)((unsigned int)buf+48)));
	printf("MAX erase count in MLC area = 0x%x !\n", *((unsigned int *)((unsigned int)buf+52)));
	printf("MIN erase count in MLC area = 0x%x !\n", *((unsigned int *)((unsigned int)buf+56)));
	printf("AVG erase count in MLC area = 0x%x !\n", *((unsigned int *)((unsigned int)buf+60)));

	printf("Number of reserved blocks for SLC area = 0x%x !\n", *((unsigned int *)((unsigned int)buf+32)));
	printf("MAX erase count in SLC area = 0x%x !\n", *((unsigned int *)((unsigned int)buf+36)));
	printf("MIN erase count in SLC area = 0x%x !\n", *((unsigned int *)((unsigned int)buf+40)));
	printf("AVG erase count in SLC area = 0x%x !\n", *((unsigned int *)((unsigned int)buf+44)));

	return 0;
}

static int mmc_get_smart_report(int dev_num)
{
	int err = 0;

#ifdef MMC_ALLOCATE_BUFFER_FROM_NONCACHE_AREA
	char * ext_csd =  non_cache_ext_csd2; // Not use non_cache_ext_csd
	if( ext_csd == NULL ) {
		printf("MMC: memory not available\n");
		return -ENOMEM;
	}
#else
	ALLOC_CACHE_ALIGN_BUFFER(char, ext_csd, 512);
#endif

	struct mmc *mmc = find_mmc_device(dev_num);
	if (!mmc){
		printf("Cann't find mmc device!\n");
		return 0;
	}

	if(IS_SD(mmc)){
		printf("SD not support smart report!\n");
		return 0;
	}

	//version 5.0
	if(mmc->version >= MMC_VERSION_5){

		//get ext_csd
		err = mmc_send_ext_csd(mmc, ext_csd);
		if(err){
			printf("Cann't read ext_csd!\n");
			return err;
		}

		//MLC-area
		printf("MLC Area(Type A):\n\t");
		switch(ext_csd[268]){
			case 0x00:
				printf("Not defined!\n");
				break;
			case 0x01:
				printf("0%% - 10%% device life time used!\n");
				break;
			case 0x02:
				printf("10%% - 20%% device life time used!\n");
				break;
			case 0x03:
				printf("20%% - 30%% device life time used!\n");
				break;
			case 0x04:
				printf("30%% - 40%% device life time used!\n");
				break;
			case 0x05:
				printf("40%% - 50%% device life time used!\n");
				break;
			case 0x06:
				printf("50%% - 60%% device life time used!\n");
				break;
			case 0x07:
				printf("60%% - 70%% device life time used!\n");
				break;
			case 0x08:
				printf("70%% - 80%% device life time used!\n");
				break;
			case 0x09:
				printf("80%% - 90%% device life time used!\n");
				break;
			case 0x0A:
				printf("90%% - 100%% device life time used!\n");
				break;
			case 0x0B:
				printf("Exceeded its maximum estimated device life time!\n");
				break;
			default:
				printf("Reserved!\n");
				break;
		}

		//SLC-area
		printf("SLC Area(Type B):\n\t");
		switch(ext_csd[269]){
			case 0x00:
				printf("Not defined!\n");
				break;
			case 0x01:
				printf("0%% - 10%% device life time used!\n");
				break;
			case 0x02:
				printf("10%% - 20%% device life time used!\n");
				break;
			case 0x03:
				printf("20%% - 30%% device life time used!\n");
				break;
			case 0x04:
				printf("30%% - 40%% device life time used!\n");
				break;
			case 0x05:
				printf("40%% - 50%% device life time used!\n");
				break;
			case 0x06:
				printf("50%% - 60%% device life time used!\n");
				break;
			case 0x07:
				printf("60%% - 70%% device life time used!\n");
				break;
			case 0x08:
				printf("70%% - 80%% device life time used!\n");
				break;
			case 0x09:
				printf("80%% - 90%% device life time used!\n");
				break;
			case 0x0A:
				printf("90%% - 100%% device life time used!\n");
				break;
			case 0x0B:
				printf("Exceeded its maximum estimated device life time!\n");
				break;
			default:
				printf("Reserved!\n");
				break;
		}

		//This field provides indication about device life time reflected by average reserved blocks.
		printf("Pre EOL info:\n\t");
		switch(ext_csd[267]){
			case 0x00:
				printf("Not defined!\n");
				break;
			case 0x01:
				printf("Normal!\n");
				break;
			case 0x02:
				printf("Warning: Consumed 80%% of reserved block!\n");
				break;
			case 0x03:
				printf("Urgent!\n");
				break;
			default:
				printf("Reserved!\n");
				break;
		}

		err = 0;
	}else{
		//other version
		switch((mmc->cid[0] >> 24) & 0xff){
			case 0x11:
				printf("Toshiba emmc:\n");
				err = toshiba_smart_report(mmc);
				break;

			case 0x15:
				printf("Samsung emmc: not support!\n");
				// not support
				err = 0;
				break;

			case 0xfe:
				printf("Micron emmc:\n");
				err = micron_smart_report(mmc);
				break;

			case 0x90:
				printf("Hynix emmc:\n");
				err = hynix_smart_report(mmc);
				break;

			default:
				printf("Unknown emmc: not support!\n");
				err = 0;
				break;
		}

	}

	return err;
}


static int mmc_set_wrt_prt(struct mmc *mmc, unsigned int addr)
{
	struct mmc_cmd cmd;
	int err;

	memset(&cmd, 0, sizeof(cmd));

	cmd.cmdidx = MMC_CMD_SET_WRITE_PROT;
	cmd.cmdarg = addr;
	cmd.resp_type = MMC_RSP_R1b;
	cmd.flags = 0;

	err = mmc_send_cmd(mmc, &cmd, NULL);
	if (err)
		return err;

	return 0;
}

static int mmc_clr_wrt_prt(struct mmc *mmc, unsigned int addr)
{
	struct mmc_cmd cmd;
	int err;

	memset(&cmd, 0, sizeof(cmd));

	cmd.cmdidx = MMC_CMD_CLR_WRITE_PROT;
	cmd.cmdarg = addr;
	cmd.resp_type = MMC_RSP_R1b;
	cmd.flags = 0;

	err = mmc_send_cmd(mmc, &cmd, NULL);
	if (err)
		return err;

	return 0;
}

static int mmc_chk_wp_status(struct mmc *mmc, unsigned int addr)
{
	struct mmc_cmd cmd;
	struct mmc_data data;
	char* tmp_buf = NULL;
	char* mmc_buf = NULL;
	int err = -1;

	tmp_buf = (char *)malloc(0x200+0x20);
	if(!tmp_buf){
		UPRINTF("%s(%d)malloc memory fail\n",__func__,__LINE__);
		goto ERR_OUT;
	}
	mmc_buf = (char *)((((uint)(tmp_buf))&(~0x1F)) + 0x20);

	memset(&cmd, 0, sizeof(cmd));
	memset(mmc_buf, 0, 0x200);

	cmd.cmdidx = MMC_CMD_SEND_WRITE_PROT;
	cmd.cmdarg = addr;
	cmd.resp_type = MMC_RSP_R1;
	cmd.flags = 0;

	data.dest = mmc_buf;
    	data.blocks = 1;
    	data.blocksize = 4;
    	data.flags = MMC_DATA_READ;

	err = mmc_send_cmd(mmc, &cmd, &data);
	if(err){
		printf("cmd %d fail!\n",cmd.cmdidx);
		goto ERR_OUT;
	}

	printf("blk 0x%08x WP status: ",cmd.cmdarg);

	{
#if 1

//     int i;

	printf("    [%02x %02x %02x %02x]\n",
		mmc_buf[0],mmc_buf[1],mmc_buf[2],mmc_buf[3]);

#else
	int i,j;

	printf("\n");
	for(i=0;i<32;i++){
		printf("    [");
		for(j=0;j<16;j++){
			printf("%02x",mmc_buf[(i*16)+j]);
		}
		printf("]\n");
	}
	printf("\n");
#endif
	}

ERR_OUT:
	if(tmp_buf)
		free(tmp_buf);

	return err;
}

static int mmc_chk_wp_type(struct mmc *mmc, unsigned int addr)
{
	struct mmc_cmd cmd;
	struct mmc_data data;
	char* tmp_buf = NULL;
	char* mmc_buf = NULL;
	int err = -1;

	tmp_buf = (char *)malloc(0x200+0x20);
	if(!tmp_buf){
		UPRINTF("%s(%d)malloc memory fail\n",__func__,__LINE__);
		goto ERR_OUT;
	}
	mmc_buf = (char *)((((uint)(tmp_buf))&(~0x1F)) + 0x20);

	memset(&cmd, 0, sizeof(cmd));
	memset(mmc_buf, 0, 0x200);

	cmd.cmdidx = MMC_CMD_SEND_WRITE_PROT_TYPE;
	cmd.cmdarg = addr;
	cmd.resp_type = MMC_RSP_R1;
	cmd.flags = 0;

	data.dest = mmc_buf;
    	data.blocks = 1;
    	data.blocksize = 8;
    	data.flags = MMC_DATA_READ;

	err = mmc_send_cmd(mmc, &cmd, &data);

	if(err){
		printf("cmd %d fail!\n",cmd.cmdidx);
		goto ERR_OUT;
	}

	printf("blk 0x%08x WP type: ",cmd.cmdarg);

	{
#if 1
	//int i;//noused
	printf("    [%02x %02x %02x %02x %02x %02x %02x %02x]\n",
		mmc_buf[0],mmc_buf[1],mmc_buf[2],mmc_buf[3],
		mmc_buf[4],mmc_buf[5],mmc_buf[6],mmc_buf[7]);

#else
	int i,j;
	printf("\n");
	for(i=0;i<32;i++){
		printf("    [");
		for(j=0;j<16;j++){
			printf("%02x",mmc_buf[(i*16)+j]);
		}
		printf("]\n");
	}
	printf("\n");
#endif
	}

ERR_OUT:
	if(tmp_buf)
		free(tmp_buf);

	return err;
}

int mmc_switch(struct mmc *mmc, u8 set, u8 index, u8 value)
{
    struct mmc_cmd cmd;
    int timeout = 1000;
    int ret;
	int no_crc = 0;

    cmd.cmdidx = MMC_CMD_SWITCH;
    cmd.resp_type = MMC_RSP_R1b;
    cmd.cmdarg = (MMC_SWITCH_MODE_WRITE_BYTE << 24) |
                 (index << 16) |
                 (value << 8) |
                 (set);
    cmd.flags = 0;

    ret = mmc_send_cmd(mmc, &cmd, NULL);
    udelay(100);

    /* Waiting for the ready status */
	if (!ret){
		if(index == EXT_CSD_HS_TIMING)
			no_crc = 1;
		ret = mmc_send_status(mmc, timeout,no_crc);
	}

    return ret;

}

int mmc_change_freq(struct mmc *mmc)
{
    char cardtype;
    int err;
#ifdef MMC_ALLOCATE_BUFFER_FROM_NONCACHE_AREA // jinn - 20140801 - fix cache line invalidate miss
    char * ext_csd =  non_cache_ext_csd2; // Not use non_cache_ext_csd
    if( ext_csd == NULL ) {
        printf("MMC: memory not available\n");
        return -ENOMEM;
    }
#else
    ALLOC_CACHE_ALIGN_BUFFER(char, ext_csd, 512);
#endif

    mmc->card_caps = 0;

    if (mmc_host_is_spi(mmc))
        return 0;

    /* Only version 4 supports high-speed */
    if (mmc->version < MMC_VERSION_4)
        return 0;

    err = mmc_send_ext_csd(mmc, ext_csd);

    if (err)
        return err;

#ifdef RTK_HS400_FUNC
    cardtype = ext_csd[EXT_CSD_CARD_TYPE];
#else
    cardtype = ext_csd[EXT_CSD_CARD_TYPE] & 0xf;
#endif

    err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL, EXT_CSD_HS_TIMING, 1);

    if (err)
        return err;

    mmc_set_bus_timing(mmc, MMC_TIMING_MMC_HS);

    /* Now check to see that it worked */
    err = mmc_send_ext_csd(mmc, ext_csd);

#ifdef RTK_HS400_FUNC
	/* get ext_csd info here */
	mmc->ext_csd.raw_card_type = ext_csd[EXT_CSD_CARD_TYPE];
	mmc->ext_csd.card_type = mmc->ext_csd.raw_card_type;
	mmc->ext_csd.driver_strength = ext_csd[EXT_CSD_DRIVER_STRENGTH];

#endif

    if (err)
        return err;

    /* No high-speed support */
    if (!ext_csd[EXT_CSD_HS_TIMING])
        return 0;

    /* High Speed is set, there are two types: 52MHz and 26MHz */
    if (cardtype & MMC_HS_52MHZ)
        mmc->card_caps |= MMC_MODE_HS_52MHz | MMC_MODE_HS;
    else
        mmc->card_caps |= MMC_MODE_HS;

#ifdef RTK_HS400_FUNC

    if (cardtype & (EXT_CSD_CARD_TYPE_SDR_1_8V|EXT_CSD_CARD_TYPE_SDR_1_2V)){

        mmc->card_caps |= MMC_MODE_HS_200MHz;

    }

    if (cardtype & (EXT_CSD_CARD_TYPE_400_1_8V|EXT_CSD_CARD_TYPE_400_1_2V)){
        mmc->card_caps |= MMC_MODE_HS_400MHz;
    }

#endif
    return 0;
}

int mmc_switch_part(int dev_num, unsigned int part_num)
{
	struct mmc *mmc = find_mmc_device(dev_num);
	u8 part_info;

	if (!mmc)
		return -1;

	part_info = (u8)((part_num & PART_ACCESS_MASK));

	return mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL, EXT_CSD_PART_CONF,
					part_info);
}

int export_mmc_switch_part(unsigned int part_num)
{
	int curr_device = 0;
	return mmc_switch_part(curr_device,part_num);
}

#if defined(Config_Module_Test_Boot_TRUE)
#define EMMC_WR_COMPARE_EN
#else
#define EMMC_WR_COMPARE_EN
#endif

#ifdef EMMC_WR_COMPARE_EN

extern int export_mmc_switch_part(unsigned int part_num);
extern int export_mmc_send_ext_csd(char *ext_csd);

//#define CHK_EXTCSD_EN
//#define DEBUG_PATTERN_EN
#define MMC_TEST_BLK_CNT 10

static int mmc_show_ext_csd(char *mmc_buf,unsigned char idx)
{
	int i;
	int ret = -1;

	memset(mmc_buf, 0, 0x200);

	ret = export_mmc_send_ext_csd(mmc_buf);
	if(ret){
		printf("mmc get ext_csd fail\n");
	}else{

		if(idx == 0xff){
			printf("ext_csd:\n");
			ret = 0;
			for(i=0;i< (512/8);i++){
				printf("[%03d]=0x%02x [%03d]=0x%02x [%03d]=0x%02x [%03d]=0x%02x "
				       "[%03d]=0x%02x [%03d]=0x%02x [%03d]=0x%02x [%03d]=0x%02x\n",
				       (i*8)+0,mmc_buf[(i*8)+0],(i*8)+1,mmc_buf[(i*8)+1],
				       (i*8)+2,mmc_buf[(i*8)+2],(i*8)+3,mmc_buf[(i*8)+3],
				       (i*8)+4,mmc_buf[(i*8)+4],(i*8)+5,mmc_buf[(i*8)+5],
				       (i*8)+6,mmc_buf[(i*8)+6],(i*8)+7,mmc_buf[(i*8)+7]);
			}
		}else{
			printf("ext_csd[%03d] = 0x%02x\n",idx,mmc_buf[idx]);
		}
	}
	return ret;
}

int export_mmc_show_ext_csd(unsigned int idx)
{
	char* tmp_buf = NULL;
	char* mmc_buf = NULL;
	int err;

	tmp_buf = (char *)malloc((MMC_TEST_BLK_CNT*0x200)+0x20);

	if(!tmp_buf){
		UPRINTF("%s(%d)malloc memory fail\n",__func__,__LINE__);
		return -1;
	}
	mmc_buf = (char *)((((uint)(tmp_buf))&(~0x1F)) + 0x20);

	err =  mmc_show_ext_csd(mmc_buf,idx);
	free(tmp_buf);

	return err;
}

#define MMC_PATTERN_1 0x00ff00ff
#define MMC_PATTERN_2 0x55aa55aa
#define MMC_PATTERN_3 0xff00ff00
#define MMC_PATTERN_4 0xaa55aa55

static void mmc_fill_pattern(char *mmc_buf,int blk_cnt)
{
	int i;
	uint *ptr;

	/*
	 change to 4 bytes, it's convenient to arrange pattern.
	 Four types pattern to test.
	 */
	 ptr = (uint*)mmc_buf;
RE_FILL:
	blk_cnt -= 1;

	for(i=0;i< 128;i++){
		if(i<32)
			ptr[i] = MMC_PATTERN_1;
		else if(i<64)
			ptr[i] = MMC_PATTERN_2;
		else if(i<96)
			ptr[i] = MMC_PATTERN_3;
		else if(i<128)
			ptr[i] = MMC_PATTERN_4;
	}
	if(blk_cnt){
		ptr += 128;
		goto RE_FILL;
	}

}

static int mmc_chk_pattern(char *mmc_buf,int blk_cnt)
{
	int i;
	int ret;
	uint *ptr;

	/*
	 change to 4 bytes, it's convenient to arrange pattern.
	 Four types pattern to test.
	 */
	ret = 0;
	ptr = (uint*)mmc_buf;

NEXT_BLK_CHK:
	blk_cnt -= 1;
	for(i=0;i< 128;i++){
		if(i<32){
			if(ptr[i] != MMC_PATTERN_1){
				ret = -1;
				break;
			}
		}
		else if(i<64){
			if(ptr[i] != MMC_PATTERN_2){
				ret = -1;
				break;
			}
		}
		else if(i<96){
			if(ptr[i] != MMC_PATTERN_3){
				ret = -1;
				break;
			}
		}
		else if(i<128){
			if(ptr[i] != MMC_PATTERN_4){
				ret = -1;
				break;
			}
		}
	}

	if(blk_cnt && !ret){
		ptr += 128;
		goto NEXT_BLK_CHK;
	}

	return ret;
}

static void mmc_show_blk(char *mmc_buf,int blk_cnt)
{
#ifdef DEBUG_PATTERN_EN

	int i;
	uint *ptr;

	ptr = (uint*)mmc_buf;

NEXT_BLK:
	blk_cnt -= 1;
	for(i=0;i<16;i++){
		printf("0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x\n",
			   ptr[(i*8)+0],ptr[(i*8)+1],ptr[(i*8)+2],ptr[(i*8)+3],
			   ptr[(i*8)+4],ptr[(i*8)+5],ptr[(i*8)+6],ptr[(i*8)+7]);
	}

	if(blk_cnt){
		ptr += 128;
		goto NEXT_BLK;
	}
#endif
}
#define EMMC_RWTEST_ADDR (0x70000)
int eMMC_W_R_compare(void)
{
	char *sor_buf = NULL;
	char *mmc_buf = NULL;
	int ret = -1;

	/* switch to boot part1 first */
//	printf("switch to boot partition 1\n");
//	ret = export_mmc_switch_part(1);
//	if(ret)
//		goto PART1;

	/* check extcsd */
	sor_buf = (char *)malloc((MMC_TEST_BLK_CNT*0x200)+0x20);
	if (sor_buf == NULL) {
		printf("mmc W/R test malloc fail\n");
		goto PART1;
	}

	mmc_buf = (char *)((((uint)(sor_buf))&(~0x1F)) + 0x20);

#ifdef CHK_EXTCSD_EN
	printf("check extcsd before\n");
	ret = mmc_show_ext_csd(mmc_buf,0xff);
	if(ret){
		goto PART1;
	}
#endif

	/* start W/R compare test  */
	memset(mmc_buf, 0, MMC_TEST_BLK_CNT*0x200);
	mmc_fill_pattern(mmc_buf,MMC_TEST_BLK_CNT);
	mmc_show_blk(mmc_buf,MMC_TEST_BLK_CNT);
	ret = rtk_eMMC_write(EMMC_RWTEST_ADDR, (MMC_TEST_BLK_CNT*0x200), (uint *)mmc_buf);
	if (ret != MMC_TEST_BLK_CNT)
	{
			/* emmc error */
			printf("mmc write fail\n");
			ret = -1;
			goto PART1;
	}

	memset(mmc_buf, 0, MMC_TEST_BLK_CNT*0x200);
	mmc_show_blk(mmc_buf,MMC_TEST_BLK_CNT);
	ret = rtk_eMMC_read(EMMC_RWTEST_ADDR, (MMC_TEST_BLK_CNT*0x200), (uint *)mmc_buf);
	if (ret != MMC_TEST_BLK_CNT)
	{
			/* emmc error */
			printf("mmc read fail\n");
			ret = -1;
			goto PART1;
	}
	mmc_show_blk(mmc_buf,MMC_TEST_BLK_CNT);
	ret = mmc_chk_pattern(mmc_buf,MMC_TEST_BLK_CNT);
	if (ret){
		printf("compare fail\n");
	}

PART1:
	//printf("switch back to user partition\n");
	//ret = export_mmc_switch_part(0) | ret;
#ifdef CHK_EXTCSD_EN
	printf("check extcsd after\n");
	ret = mmc_show_ext_csd(mmc_buf,0xff);
#endif

	if(sor_buf)
		free(sor_buf);
	return ret;
}

int export_eMMC_W_R_compare(void){
	eMMC_W_R_compare();
	return 0;
}

int eMMC_W_R_compare_test(unsigned int blk_adr,unsigned int blk_cnt_req,unsigned int loop)
{
	char *sor_buf = NULL;
	char *mmc_buf = NULL;
	int ret = -1;
	int bb = 0;
	int i;
	unsigned int blk_cnt = blk_cnt_req;

RE_MALLOC:
	sor_buf = (char *)malloc((blk_cnt*0x200)+0x20);
	if (sor_buf == NULL) {

		if(blk_cnt > 100){
			blk_cnt -= 100;
			bb = 1;
			goto RE_MALLOC;
		}else{
			printf("mmc W/R test malloc %dblk buff fail\n",blk_cnt);
			goto ERR_OUT;
		}
	}
	if(bb)
		printf("modify testing blk to %dblk\n",blk_cnt);
	printf("testing blk cnt is %dblk\n",blk_cnt);

	mmc_buf = (char *)((((uint)(sor_buf))&(~0x1F)) + 0x20);

	for(i=0;i<loop;i++){
		/* start W/R compare test  */
		memset(mmc_buf, 0, blk_cnt*0x200);
		mmc_fill_pattern(mmc_buf,blk_cnt);
		mmc_show_blk(mmc_buf,blk_cnt);
		ret = rtk_eMMC_write(blk_adr, (blk_cnt*0x200), (uint *)mmc_buf);
		if (ret != blk_cnt)
		{
			/* emmc error */
			printf("mmc write fail at loop %d/%d\n",i,loop);
			ret = -1;
			break;
		}

		memset(mmc_buf, 0, blk_cnt*0x200);
		mmc_show_blk(mmc_buf,blk_cnt);
		ret = rtk_eMMC_read(blk_adr, (blk_cnt*0x200), (uint *)mmc_buf);
		if (ret != blk_cnt)
		{
			/* emmc error */
			printf("mmc read fail at loop %d/%d\n",i,loop);
			ret = -1;
			break;
		}
		mmc_show_blk(mmc_buf,blk_cnt);
		ret = mmc_chk_pattern(mmc_buf,blk_cnt);
		if (ret){
			printf("compare fail at loop %d/%d\n",i,loop);
			break;
		}
		if((i%100) == 0)
			printf("compare pass after loop %d/%d\n",i,loop);
	}

ERR_OUT:
	if(sor_buf)
		free(sor_buf);

	return ret;
}

int export_eMMC_W_R_compare_test(unsigned int blk_adr,
		unsigned int blk_cnt,unsigned int loop)
{
	int ret = 0;
	ret = eMMC_W_R_compare_test(blk_adr,blk_cnt,loop);
	return ret;
}

int export_eMMC_W_pattern(void)
{
	printf("%s(%d)no support yet!\n",__func__,__LINE__);
	return 0;
}

int export_eMMC_chk_pattern(void)
{
	printf("%s(%d)no support yet!\n",__func__,__LINE__);
	return 0;
}
#endif

static int mmc_wp_manage(
	int dev_num,
	unsigned long addr,
	unsigned long blkcnt,
	unsigned char wp_en)
{
	int err = 0;
	unsigned int blk = 0;
	unsigned int wp_grp_sz = 0;
#if 0
	int set_val = 0;
#endif
#ifdef MMC_ALLOCATE_BUFFER_FROM_NONCACHE_AREA
	char * ext_csd =  non_cache_ext_csd2; // Not use non_cache_ext_csd
	if( ext_csd == NULL ) {
		printf("MMC: memory not available\n");
		return -ENOMEM;
	}
#else
	ALLOC_CACHE_ALIGN_BUFFER(char, ext_csd, 512);
#endif

	struct mmc *mmc = find_mmc_device(dev_num);
	if (!mmc){
		printf("Cann't find mmc device!\n");
		return -ENODEV;
	}

	// sanity check
	if(IS_SD(mmc)){
		printf("SD not support wp set!\n");
		return -1;
	}

	if(blkcnt == 0){
		printf("argument blkcnt setting error!\n");
		return -1;
	}

	if ((addr + blkcnt) > mmc->block_dev.lba) {
		printf("MMC: block number 0x%lx exceeds max(0x%lx)\n",
			addr + blkcnt, mmc->block_dev.lba);
		return -1;
	}

	//get ext_csd
	err = mmc_send_ext_csd(mmc, ext_csd);
	if(err){
		printf("Cann't read ext_csd!\n");
		return -1;
	}

	if(wp_en == WP_CHECK_INFO){
		UPRINTF("\n"
			"Manufacturer ID: %x\n", mmc->cid[0] >> 24);
		UPRINTF("OEM: %x\n", (mmc->cid[0] >> 8) & 0xffff);
		UPRINTF("Name: %c%c%c%c%c\n\n", (mmc->cid[0] & 0xff),
			(mmc->cid[1] >> 24) & 0xff,
			(mmc->cid[1] >> 16) & 0xff,
			(mmc->cid[1] >> 8) & 0xff,
			(mmc->cid[1] >> 0) & 0xff);

		UPRINTF("CSD[raw]: %08x%08x%08x%08x\n",
			mmc->csd[0],mmc->csd[1],mmc->csd[2],mmc->csd[3]);

		UPRINTF("CSD[12] = %d (TMP_WRITE_PROTECT)\n"
			"CSD[13] = %d (PERM_WRITE_PROTECT)\n\n",
			(mmc->csd[3] >> 12) & 0x01,
			(mmc->csd[3] >> 13) & 0x01);

		printf( "ECSD[ 59] = 0x%02x (CLASS_6_CTRL)\n"
			"ECSD[175] = 0x%02x (ERASE_GROUP_DEF)\n"
			"ECSD[171] = 0x%02x (USER_WP)\n"
			"ECSD[221] = 0x%02x (HC_WP_GRP_SIZE)\n"
			"ECSD[224] = 0x%02x (HC_ERASE_GRP_SIZE)\n"
			"ECSD[173] = 0x%02x (BOOT_WP)\n"
			"ECSD[174] = 0x%02x (BOOT_WP_STATUS)\n"
			"ECSD[178] = 0x%02x (BOOT_CONFIG_PROT)\n\n",
			ext_csd[59],ext_csd[175],
			ext_csd[171],ext_csd[221],ext_csd[224],
			ext_csd[173],ext_csd[174],ext_csd[178]);

		printf( "Write Protect Group Size\n"
			"       = (512KB*HC_ERASE_GRP_SIZE)*HC_WP_GRP_SIZE\n"
			"       = 0x%x (blocks)\n"
			"Total Block Count is 0x%08x\n",
			((1024*ext_csd[224])*ext_csd[221]),
			((ext_csd[EXT_CSD_SEC_CNT] << 0) |
			 (ext_csd[EXT_CSD_SEC_CNT + 1] << 8) |
			 (ext_csd[EXT_CSD_SEC_CNT + 2] << 16) |
			 (ext_csd[EXT_CSD_SEC_CNT + 3] << 24)));
		return 0;
	}



	// check class_6_ctrl
	if(ext_csd[59] != 0x0){
		printf("Dynamic capacity is enable\n");
		return -1;
	}

	// set ext_csd[175] 0x1
	err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL, 175, 0x1);
	if(err < 0){
		printf("set ERASE_GROUP_DEF bit0 fail\n");
		return -1;
	}else{
		mmc_send_status(mmc, 1000,0);
	}

	wp_grp_sz = 1024 * ext_csd[224] * ext_csd [221]; // 512KB * HC_ERASE_GRP_SIZE * HC_WP_GRP_SIZE

	if ((addr % wp_grp_sz) || (blkcnt % wp_grp_sz)) {
		printf("\n\nCaution! Your device's wp group size is 0x%x\n"
				"The start address should aligned to 0x%lx. "
				"blkcnt 0x%lx should be multiple of  wp_grp_sz\n\n",
				wp_grp_sz, addr & ~(wp_grp_sz - 1), blkcnt);
		return -1;
	}

#if 0
	set_val = 0x0;//| ext_csd[171];
	err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL, 171, set_val);
	if (err < 0) {
		printf("cancel permanent write protect fail\n");
		return -1;
	}else{
		mmc_send_status(mmc, 1000,0);
	}

	err = mmc_send_ext_csd(mmc, ext_csd);
	if (err < 0) {
		printf("send cmd8 fail\n");
		return -1;
	}
#endif

	if ((wp_en == WP_DISABLE) || (wp_en == WP_ENABLE)){
		printf("write %s from 0x%08lx to 0x%08lx\n",
			(wp_en == WP_ENABLE)? "set protect":"clear protect", addr, addr + blkcnt - 1);
		for (blk=0; blk < blkcnt; blk += wp_grp_sz) {
			if (wp_en == WP_DISABLE){
				err = mmc_clr_wrt_prt(mmc, addr + blk);
			}else{
				err = mmc_set_wrt_prt(mmc, addr + blk);
			}
			if (err) {
				return err;
			}
			if (mmc_send_status(mmc, 10000,0))
				return -1;
		}

	}else if( (wp_en == WP_CHECK_STAT)
		||(wp_en == WP_CHECK_TYPE)){
		printf("check write protect %s from 0x%08lx to 0x%08lx\n",
			(wp_en == WP_CHECK_STAT)?
			"status":"type", addr, addr + blkcnt - 1);
		/* One chk command can display information about 32 write protect group */
		for (blk=0; blk < blkcnt; blk += wp_grp_sz*32) {
			if(wp_en == WP_CHECK_STAT){
				/* action to check protect information */
				err = mmc_chk_wp_status(mmc, addr + blk);
			}else{
				/* action to check protect type */
				err = mmc_chk_wp_type(mmc, addr + blk);
			}
			if (err) {
				return err;
			}
			if (mmc_send_status(mmc, 10000,0))
				return -1;
		}

		if(wp_en == WP_CHECK_TYPE){
			printf(	"    '00' Write protection group not protected\n"
				"    '01' Write protection group is protected by temporary write protection\n"
				"    '10' Write protection group is protected by power-on write protection\n"
				"    '11' Write protection group is protected by permanent write protection\n");
		}
	}else{
		printf("unknown operation mode\n");
		return -1;
	}

	return 0;
}

int mmc_getcd(struct mmc *mmc)
{
    int cd;

    cd = board_mmc_getcd(mmc);

    if ((cd < 0) && mmc->getcd)
        cd = mmc->getcd(mmc);

    return cd;
}

int sd_switch(struct mmc *mmc, int mode, int group, u8 value, u8 *resp)
{
    struct mmc_cmd cmd;
    struct mmc_data data;

    /* Switch the frequency */
    cmd.cmdidx = SD_CMD_SWITCH_FUNC;
    cmd.resp_type = MMC_RSP_R1;
    cmd.cmdarg = (mode << 31) | 0xffffff;
    cmd.cmdarg &= ~(0xf << (group * 4));
    cmd.cmdarg |= value << (group * 4);
    cmd.flags = 0;

    data.dest = (char *)resp;
    data.blocksize = 64;
    data.blocks = 1;
    data.flags = MMC_DATA_READ;

    return mmc_send_cmd(mmc, &cmd, &data);
}


int sd_change_freq(struct mmc *mmc)
{
    int err;
    struct mmc_cmd cmd;
    ALLOC_CACHE_ALIGN_BUFFER(uint, scr, 2);
    ALLOC_CACHE_ALIGN_BUFFER(uint, switch_status, 16);
    struct mmc_data data;
    int timeout;

    mmc->card_caps = 0;

    if (mmc_host_is_spi(mmc))
        return 0;

    /* Read the SCR to find out if this card supports higher speeds */
    cmd.cmdidx = MMC_CMD_APP_CMD;
    cmd.resp_type = MMC_RSP_R1;
    cmd.cmdarg = mmc->rca << 16;
    cmd.flags = 0;

    err = mmc_send_cmd(mmc, &cmd, NULL);

    if (err)
        return err;

    cmd.cmdidx = SD_CMD_APP_SEND_SCR;
    cmd.resp_type = MMC_RSP_R1;
    cmd.cmdarg = 0;
    cmd.flags = 0;

    timeout = 3;

retry_scr:
    data.dest = (char *)scr;
    data.blocksize = 8;
    data.blocks = 1;
    data.flags = MMC_DATA_READ;

    err = mmc_send_cmd(mmc, &cmd, &data);

    if (err) {
        if (timeout--)
            goto retry_scr;

        return err;
    }

    mmc->scr[0] = __be32_to_cpu(scr[0]);
    mmc->scr[1] = __be32_to_cpu(scr[1]);

    switch ((mmc->scr[0] >> 24) & 0xf) {
        case 0:
            mmc->version = SD_VERSION_1_0;
            break;
        case 1:
            mmc->version = SD_VERSION_1_10;
            break;
        case 2:
            mmc->version = SD_VERSION_2;
            break;
        default:
            mmc->version = SD_VERSION_1_0;
            break;
    }

    if (mmc->scr[0] & SD_DATA_4BIT)
        mmc->card_caps |= MMC_MODE_4BIT;

    /* Version 1.0 doesn't support switching */
    if (mmc->version == SD_VERSION_1_0)
        return 0;

    timeout = 4;
    while (timeout--) {
        err = sd_switch(mmc, SD_SWITCH_CHECK, 0, 1,
                (u8 *)switch_status);

        if (err)
            return err;

        /* The high-speed function is busy.  Try again */
        if (!(__be32_to_cpu(switch_status[7]) & SD_HIGHSPEED_BUSY))
            break;
    }

    /* If high-speed isn't supported, we return */
    if (!(__be32_to_cpu(switch_status[3]) & SD_HIGHSPEED_SUPPORTED))
        return 0;

    /*
     * If the host doesn't support SD_HIGHSPEED, do not switch card to
     * HIGHSPEED mode even if the card support SD_HIGHSPPED.
     * This can avoid furthur problem when the card runs in different
     * mode between the host.
     */
    if (!((mmc->host_caps & MMC_MODE_HS_52MHz) &&
        (mmc->host_caps & MMC_MODE_HS)))
        return 0;

    err = sd_switch(mmc, SD_SWITCH_SWITCH, 0, 1, (u8 *)switch_status);

    if (err)
        return err;

    if ((__be32_to_cpu(switch_status[4]) & 0x0f000000) == 0x01000000)
        mmc->card_caps |= MMC_MODE_HS;

    return 0;
}

/* frequency bases */
/* divided by 10 to be nice to platforms without floating point */
static const int fbase[] = {
    10000,
    100000,
    1000000,
    10000000,
};

/* Multiplier values for TRAN_SPEED.  Multiplied by 10 to be nice
 * to platforms without floating point.
 */
static const int multipliers[] = {
    0,  /* reserved */
    10,
    12,
    13,
    15,
    20,
    25,
    30,
    35,
    40,
    45,
    50,
    55,
    60,
    70,
    80,
};

void mmc_set_ios(struct mmc *mmc)
{
    mmc->set_ios(mmc);
}

void mmc_set_clock(struct mmc *mmc, uint clock)
{
    if (clock > mmc->f_max)
        clock = mmc->f_max;

    if (clock < mmc->f_min)
        clock = mmc->f_min;

    mmc->clock = clock;

    mmc_set_ios(mmc);
}

void mmc_set_bus_width(struct mmc *mmc, uint width)
{
    mmc->bus_width = width;

    mmc_set_ios(mmc);
}

void mmc_set_bus_timing(struct mmc *mmc, uint timing)
{
    mmc->timing = timing;

    mmc_set_ios(mmc);
}

void mmc_hexdump(const void *buf, unsigned int length)
{
    unsigned int i;
    unsigned char *ptr = (unsigned char *)buf;

    if ((buf == NULL) || (length == 0))
    {
        printf("NULL\n");
        return;
    }

    printf("===> %s(0x%x, 0x%x):\n", __FUNCTION__, (unsigned int)buf, length);
    for (i = 0; i < length; i++)
    {
        if( (i & 0x0f) == 0 ) {
            printf("%03d :", i);
        }
        printf(" %02X", ptr[i]);

        if ((i & 0xf) == 0xf)
        {
            printf("\n");
        }
    }
    printf("\n");
}


#ifdef RTK_HS400_FUNC
static int mmc_select_hs200(struct mmc *mmc)
{
	int err = 0;
	e_device_type* sdport = (e_device_type*)mmc->priv;
	unsigned int mid = sdport->mid;
	unsigned int ds_select = 0;	// driver strength selected

	//all device use 0x4(40ohm:x1.2) now
	ds_select = (mid == 0x11)? 0x4: 0x4;

	/*
 	 At kernel mmc_select_hs200(), there are check voltage 1.8 or 1.2.
 	 RTK fix at 1.8, skip this./
	 */

	/*
	Notice :
	 1. Hs200 runing on 4/8 bit only, check this first
	 2. In here, we sould to know what bus width now. to tuning on 4 or 8 bit
	 */
	/* device driving strangth depend on mmc->ext_csd.driver_strength */
	err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL,
		 	EXT_CSD_HS_TIMING,
			EXT_CSD_TIMING_HS200|(ds_select<<4));

	printf("HS200 ");
	if(err){
		printf("fail!!!\n");
	} else {
		printf("enable\n");
		mmc_set_bus_timing(mmc, MMC_TIMING_MMC_HS200);
	}

	return err;

}

static int mmc_select_hs400(struct mmc *mmc)
{
	int err = 0;
	e_device_type* sdport = (e_device_type*)mmc->priv;
	unsigned int mid = sdport->mid;
	unsigned int ds_select = 0;	// driver strength selected

	//all device use 0x4(40ohm:x1.2) now
	ds_select = (mid == 0x11)? 0x4: 0x4;

	/*
 	 At kernel mmc_select_hs400(), there are check voltage 1.8 or 1.2.
 	 RTK fix at 1.8, skip this./
	 */
	/*
	Notice :
	 1. Hs400 runing on 8 bit only, we check before.
	 2. down speed mode to high-speed first
	 */

	mmc->tran_speed = 50000000;
	mmc_set_clock(mmc, mmc->tran_speed);
	err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL, EXT_CSD_HS_TIMING, 1);

	if(err){
		printf("mmc: back to high speed mode fail.\n");
		/*  Should try more time ??*/
		return err;
	}
	mmc_set_bus_timing(mmc, MMC_TIMING_SD_HS);
	mmc->tran_speed = 50000000;
	mmc_set_clock(mmc, mmc->tran_speed);

	err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL,
			EXT_CSD_BUS_WIDTH,
			EXT_CSD_DDR_BUS_WIDTH_8);
	if (err) {
		printf("%mmc: switch to bus width for hs400 failed\n");
		return err;
	}

	/* device driving strangth depend on mmc->ext_csd.driver_strength */
	err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL,
		 	EXT_CSD_HS_TIMING,
		 	EXT_CSD_TIMING_HS400|(ds_select<<4));

	printf("HS400 ");
	if(err){
		printf("fail!!!\n");
	} else {
		printf("enable\n");
		mmc_set_bus_timing(mmc, MMC_TIMING_MMC_HS400);
		mmc->tran_speed = 200000000;
		mmc_set_clock(mmc, mmc->tran_speed);
	}

	return err;
}

#endif

extern void export_mmc_print_ext_csd(char *ext_csd);
int mmc_startup(struct mmc *mmc)
{
	int err, width;
	uint mult, freq;
	u64 cmult, csize, capacity;
	struct mmc_cmd cmd;
	int timeout = 1000;

#define CSD_ARRAY_SIZE 512
#ifdef MMC_ALLOCATE_BUFFER_FROM_NONCACHE_AREA // jinn - 20140801 - fix cache line invalidate miss
	char * ext_csd = non_cache_ext_csd;
	char * test_csd = non_cache_ext_csd2;
	if( ext_csd == NULL || test_csd == NULL ) {
		printf("MMC: memory not available\n");
		return -ENOMEM;
	}
#else
	ALLOC_CACHE_ALIGN_BUFFER(char, ext_csd, CSD_ARRAY_SIZE);
	ALLOC_CACHE_ALIGN_BUFFER(char, test_csd, CSD_ARRAY_SIZE);
#endif

#ifdef CONFIG_MMC_SPI_CRC_ON
	if (mmc_host_is_spi(mmc)) { /* enable CRC check for spi */
		cmd.cmdidx = MMC_CMD_SPI_CRC_ON_OFF;
		cmd.resp_type = MMC_RSP_R1;
		cmd.cmdarg = 1;
		cmd.flags = 0;
		err = mmc_send_cmd(mmc, &cmd, NULL);

		if (err)
			return err;
	}
#endif

	/* Put the Card in Identify Mode */
	cmd.cmdidx = mmc_host_is_spi(mmc) ?
		MMC_CMD_SEND_CID : MMC_CMD_ALL_SEND_CID; /* cmd not supported in spi */
	cmd.resp_type = MMC_RSP_R2;
	cmd.cmdarg = 0;
	cmd.flags = 0;

	err = mmc_send_cmd(mmc, &cmd, NULL);

	if (err)
		return err;

	memcpy(mmc->cid, cmd.response, 16);

	/*
	* For MMC cards, set the Relative Address.
	* For SD cards, get the Relatvie Address.
	* This also puts the cards into Standby State
	*/
	if (!mmc_host_is_spi(mmc)) { /* cmd not supported in spi */
		cmd.cmdidx = SD_CMD_SEND_RELATIVE_ADDR;
		cmd.cmdarg = mmc->rca << 16;
		cmd.resp_type = MMC_RSP_R6;
		cmd.flags = 0;

		err = mmc_send_cmd(mmc, &cmd, NULL);

		if (err)
			return err;

		if (IS_SD(mmc)){
			printf("SD fake eMMC detected!\n");
			mmc->rca = (cmd.response[0] >> 16) & 0xffff;
		}
	}

	/* Get the Card-Specific Data */
	cmd.cmdidx = MMC_CMD_SEND_CSD;
	cmd.resp_type = MMC_RSP_R2;
	cmd.cmdarg = mmc->rca << 16;
	cmd.flags = 0;

	err = mmc_send_cmd(mmc, &cmd, NULL);

	/* Waiting for the ready status */
	mmc_send_status(mmc, timeout,0);

	if (err)
		return err;

	mmc->csd[0] = cmd.response[0];	//127:96
	mmc->csd[1] = cmd.response[1];	// 95:62
	mmc->csd[2] = cmd.response[2];	// 63:32
	mmc->csd[3] = cmd.response[3];	// 31: 0

	if (mmc->version == MMC_VERSION_UNKNOWN) {
		int version = (cmd.response[0] >> 26) & 0xf;

		switch (version) {
		case 0:
			mmc->version = MMC_VERSION_1_2;
			break;
		case 1:
			mmc->version = MMC_VERSION_1_4;
			break;
		case 2:
			mmc->version = MMC_VERSION_2_2;
			break;
		case 3:
			mmc->version = MMC_VERSION_3;
			break;
		case 4:
			mmc->version = MMC_VERSION_4;
			break;
		default:
			mmc->version = MMC_VERSION_1_2;
			break;
		}
	}

	/* divide frequency by 10, since the mults are 10x bigger */
	freq = fbase[(cmd.response[0] & 0x7)];
	mult = multipliers[((cmd.response[0] >> 3) & 0xf)];

	mmc->tran_speed = freq * mult;

	mmc->read_bl_len = 1 << ((cmd.response[1] >> 16) & 0xf);

	if (IS_SD(mmc))
		mmc->write_bl_len = mmc->read_bl_len;
	else
		mmc->write_bl_len = 1 << ((cmd.response[3] >> 22) & 0xf);

	if (mmc->high_capacity) {
		csize = (mmc->csd[1] & 0x3f) << 16
			| (mmc->csd[2] & 0xffff0000) >> 16;
		cmult = 8;
	} else {
		csize = (mmc->csd[1] & 0x3ff) << 2
			| (mmc->csd[2] & 0xc0000000) >> 30;
		cmult = (mmc->csd[2] & 0x00038000) >> 15;
	}

	mmc->capacity = (csize + 1) << (cmult + 2);
	mmc->capacity *= mmc->read_bl_len;

	if (mmc->read_bl_len > 512)
		mmc->read_bl_len = 512;

	if (mmc->write_bl_len > 512)
		mmc->write_bl_len = 512;

	/* Select the card, and put it into Transfer Mode */
	if (!mmc_host_is_spi(mmc)) { /* cmd not supported in spi */
		cmd.cmdidx = MMC_CMD_SELECT_CARD;
		cmd.resp_type = MMC_RSP_R1;
		cmd.cmdarg = mmc->rca << 16;
		cmd.flags = 0;
		err = mmc_send_cmd(mmc, &cmd, NULL);

		if (err)
			return err;
	}

	/* enter pullpush state, set clock big then 400kHz */
	//mmc_set_clock(mmc, 400001);

	/*
	* For SD, its erase group is always one sector
	*/
	mmc->erase_grp_size = 1;
	mmc->part_config = MMCPART_NOAVAILABLE;
    	if (!IS_SD(mmc) && (mmc->version >= MMC_VERSION_4)) {
		/* check  ext_csd version and capacity */
#ifndef MMC_ALLOCATE_BUFFER_FROM_NONCACHE_AREA
		flush_cache((unsigned long)ext_csd, CSD_ARRAY_SIZE);
#endif
		err = mmc_send_ext_csd(mmc, ext_csd);

		if(!err){
			e_device_type *host = (e_device_type *)mmc->priv;
			memcpy(host->raw_ecsd, ext_csd, CSD_ARRAY_SIZE);
		}

		if(ext_csd[EXT_CSD_REV] < 7){
			mmc->version = MMC_VERSION_4 | (ext_csd[EXT_CSD_REV]);
		}else{
			mmc->version = MMC_VERSION_5 | (ext_csd[EXT_CSD_REV]-7);
		}

		//V1 clockworks warning - fix ID 816/835
		if(!err){
			// init ERASED_MEM_CONT
			mmc->ext_csd.raw_erased_mem_count = ext_csd[EXT_CSD_ERASED_MEM_CONT];
		}

		if (!err & (ext_csd[EXT_CSD_REV] >= 2)) {
			/*
			* According to the JEDEC Standard, the value of
			* ext_csd's capacity is valid if the value is more
			* than 2GB
			*/
			//mmc_hexdump(ext_csd, CSD_ARRAY_SIZE);
			capacity = ext_csd[EXT_CSD_SEC_CNT] << 0
				 | ext_csd[EXT_CSD_SEC_CNT + 1] << 8
				 | ext_csd[EXT_CSD_SEC_CNT + 2] << 16
				 | ext_csd[EXT_CSD_SEC_CNT + 3] << 24;
			capacity = capacity*512;
			if ((capacity >> 20) > 2 * 1024)
				mmc->capacity = capacity;
		}

		/*
		 * Check whether GROUP_DEF is set, if yes, read out
		 * group size from ext_csd directly, or calculate
		 * the group size from the csd value.
		 */
		if (ext_csd[EXT_CSD_ERASE_GROUP_DEF])
			mmc->erase_grp_size =
				ext_csd[EXT_CSD_HC_ERASE_GRP_SIZE] * 512 * 1024;
		else {
			int erase_gsz, erase_gmul;
			erase_gsz = (mmc->csd[2] & 0x00007c00) >> 10;
			erase_gmul = (mmc->csd[2] & 0x000003e0) >> 5;
			mmc->erase_grp_size = (erase_gsz + 1)
						* (erase_gmul + 1);
		}

		/* store the partition info of emmc */
		if (ext_csd[EXT_CSD_PARTITIONING_SUPPORT] & PART_SUPPORT)
			mmc->part_config = ext_csd[EXT_CSD_PART_CONF];

		mmc->sec_support = ext_csd[EXT_CSD_SEC_FEATURE_SUPPORT];

	}

	export_mmc_print_ext_csd(ext_csd);

	if (IS_SD(mmc))
		err = sd_change_freq(mmc);
	else
		err = mmc_change_freq(mmc);

	if (err)
		return err;

	/* Restrict card's capabilities by what the host can do */
	mmc->card_caps &= mmc->host_caps;
	width = 0;
	if (IS_SD(mmc)) {
		if (mmc->card_caps & MMC_MODE_4BIT) {
			cmd.cmdidx = MMC_CMD_APP_CMD;
			cmd.resp_type = MMC_RSP_R1;
			cmd.cmdarg = mmc->rca << 16;
			cmd.flags = 0;

			err = mmc_send_cmd(mmc, &cmd, NULL);
			if (err)
				return err;

			cmd.cmdidx = SD_CMD_APP_SET_BUS_WIDTH;
			cmd.resp_type = MMC_RSP_R1;
			cmd.cmdarg = 2;
			cmd.flags = 0;
			err = mmc_send_cmd(mmc, &cmd, NULL);
			if (err)
				return err;

			mmc_set_bus_width(mmc, 4);
		}

		if (mmc->card_caps & MMC_MODE_HS)
			mmc->tran_speed = 50000000;
		else
			mmc->tran_speed = 25000000;

	} else {
		width = (mmc->host_caps & MMC_MODE_MASK_WIDTH_BITS);

		if(width & MMC_MODE_8BIT){
			width = MMC_MODE_8BIT;
		}else if(width & MMC_MODE_4BIT){
			width = MMC_MODE_4BIT;
		}else{
			width = 0;
		}
		width = (width>>MMC_MODE_WIDTH_BITS_SHIFT);

		for (; width >= 0; width--) {
			/* Set the card to use 4 bit*/
			err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL,
			EXT_CSD_BUS_WIDTH, width);

			if (err)
				continue;

			if (!width) {
				mmc_set_bus_width(mmc, 1);
				break;
			} else
				mmc_set_bus_width(mmc, 4 * width);

#ifndef MMC_ALLOCATE_BUFFER_FROM_NONCACHE_AREA
			flush_cache((unsigned long)test_csd, CSD_ARRAY_SIZE);
#endif

RE_SET_TYPE:
			err = mmc_send_ext_csd(mmc, test_csd);

			if (!err \
			 && ext_csd[EXT_CSD_PARTITIONING_SUPPORT] \
			   == test_csd[EXT_CSD_PARTITIONING_SUPPORT]
			 && ext_csd[EXT_CSD_ERASE_GROUP_DEF] \
			   == test_csd[EXT_CSD_ERASE_GROUP_DEF] \
			 && ext_csd[EXT_CSD_REV] \
			   == test_csd[EXT_CSD_REV]
			 && ext_csd[EXT_CSD_HC_ERASE_GRP_SIZE] \
			   == test_csd[EXT_CSD_HC_ERASE_GRP_SIZE]
			 && memcmp(&ext_csd[EXT_CSD_SEC_CNT], \
			   &test_csd[EXT_CSD_SEC_CNT], 4) == 0)
			{
#ifdef RTK_HS400_FUNC
				/*
				Here, record current runing bus width but not capability of card.
				mmc->bus_width is same meaning but for driver.
				*/
				if(width == 2){
					mmc->card_caps |= (MMC_MODE_8BIT|MMC_MODE_4BIT);
				}else if(width == 1){
					mmc->card_caps |= (MMC_MODE_4BIT);
				}
#else
				mmc->card_caps |= width;
#endif
				break;
			}
			else {
				if (!width) {
					printf("MMC: set bus width 1 fail\n");
				}
				else {
					printf("MMC: set bus width %d fail, try other mode\n", (4 * width));
				}
			}
		}

		if (mmc->card_caps & MMC_MODE_HS) {
			if (mmc->card_caps & MMC_MODE_HS_52MHz)
				mmc->tran_speed = 52000000;
			else
				mmc->tran_speed = 26000000;
		}
	}

#ifdef RTK_HS400_FUNC
	/* check HS200 */
	if (mmc->card_caps & MMC_MODE_HS_200MHz){
		if(mmc->card_caps & ((MMC_MODE_8BIT|MMC_MODE_4BIT))){

			err = mmc_select_hs200(mmc);
			if(!err){
				mmc->tran_speed = 200000000;
				mmc_set_clock(mmc, mmc->tran_speed);

				/*
		 		 * For devices supporting HS200 mode, the bus width has
				 * to be set before executing the tuning function. If
		 		 * set before tuning, then device will respond with CRC
		 		 * errors for responses on CMD line. So for HS200 the
		  		 * sequence will be
		 		 * 1. set bus width 4bit / 8 bit (1 bit not supported)
		 		 * 2. switch to HS200 mode
		 		 * 3. set the clock to > 52Mhz <=200MHz and
		 		 * 4. execute tuning for HS200
		 		 */
				err = mmc->execute_tuning(mmc,21);

				if (err) {
#if (defined(Config_Module_Test_Boot_TRUE) || defined(Config_FW_Type_FT))
					/* skip go back to HS50 mode to verify IC in quality  */
					return err;
#else
					/*
					 if tuning, It's need to retry to result of system keep runing
					 1. clear HS200 & HS400 flag
					 2. back to High Speed mode
					 */
					printf("mmc: hs200 tuning execution failed\n");

					mmc->host_caps &= ~(MMC_MODE_HS_200MHz|MMC_MODE_HS_400MHz);
					mmc->card_caps &= mmc->host_caps;

					/* HS200 but down speed to 50Mhz */
					mmc->tran_speed = 50000000;
					mmc_set_clock(mmc, mmc->tran_speed);
					err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL, EXT_CSD_HS_TIMING, 1);

					mmc_set_bus_timing(mmc, MMC_TIMING_MMC_HS);

					if (err){
						printf("mmc: back to high speed mode fail.\n");
						/*  Should try more time ??*/
						return err;
					}
					goto RE_SET_TYPE;
#endif
				}
			}
		}else{
			printf("%s(%d)have not runing on 4 or 8 bit, skip HS200 mode\n",
				__func__,__LINE__);
		}
	}else{
		mmc_set_clock(mmc, mmc->tran_speed);
	}

	if ((mmc->card_caps & MMC_MODE_HS_200MHz)
	 && (mmc->card_caps & MMC_MODE_HS_400MHz)
	 && (mmc->card_caps & MMC_MODE_8BIT))
	{
		err = mmc_select_hs400(mmc);
		if(!err){
			/*
	 		 * For devices supporting HS400 mode, the bus width has
			 * to be 8 bit .
			 * So for HS400 the  sequence will be
			 * 1. runing on HS200 and 8-bit
			 * 2. switch to High-Speed mode(50Mhz)
	 		 * 3. switch to HS400 mode
	 		 */
			err = mmc->execute_tuning(mmc,18);

			if (err) {
				/*
				 if tuning, It's need to retry to result of system keep runing
				 1. clear HS200 & HS400 flag
				 2. back to High Speed mode
				 */
				printf("mmc: hs400 tuning execution failed\n");

				mmc->host_caps &= ~(MMC_MODE_HS_400MHz);
				mmc->card_caps &= mmc->host_caps;

				/* HS400 but down speed to 50Mhz */
				mmc->tran_speed = 50000000;
				mmc_set_clock(mmc, mmc->tran_speed);

				err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL, EXT_CSD_HS_TIMING, 1);
				mmc_set_bus_timing(mmc, MMC_TIMING_MMC_HS);

				if(err){
					printf("mmc: back to high speed mode fail.\n");
					/*  Should try more time ??*/
					return err;
				}
				goto RE_SET_TYPE;
			}
		}else{
			//HS400 fail
			mmc_send_status(mmc, 1000,1);
		}

	}

#endif

	/* fill in device description */
	mmc->block_dev.lun = 0;
	mmc->block_dev.type = 0;
	mmc->block_dev.blksz = mmc->read_bl_len;
	mmc->block_dev.lba = lldiv(mmc->capacity, mmc->read_bl_len);
	sprintf(mmc->block_dev.vendor, "Man %06x Snr %08x", mmc->cid[0] >> 8,
			(mmc->cid[2] << 8) | (mmc->cid[3] >> 24));
	sprintf(mmc->block_dev.product, "%c%c%c%c%c", mmc->cid[0] & 0xff,
			(mmc->cid[1] >> 24), (mmc->cid[1] >> 16) & 0xff,
			(mmc->cid[1] >> 8) & 0xff, mmc->cid[1] & 0xff);
	sprintf(mmc->block_dev.revision, "%d.%d", mmc->cid[2] >> 28,
			(mmc->cid[2] >> 24) & 0xf);
	init_part(&mmc->block_dev);
#ifdef CONFIG_QC_VERIFY
        //err = eMMC_W_R_compare();
        err = export_eMMC_W_R_compare_test(0x70000, 450, 1);
        if (err) {
            printf("W_R_compare test fail\n");
            return err;
        }
        else
            printf("W_R_compare test pass\n");
#endif
	return 0;
}

int mmc_send_if_cond(struct mmc *mmc)
{
    struct mmc_cmd cmd;
    int err;

    cmd.cmdidx = SD_CMD_SEND_IF_COND;
    /* We set the bit if the host supports voltages between 2.7 and 3.6 V */
    cmd.cmdarg = ((mmc->voltages & 0xff8000) != 0) << 8 | 0xaa;
    cmd.resp_type = MMC_RSP_R7;
    cmd.flags = 0;

    err = mmc_send_cmd(mmc, &cmd, NULL);

    if (err)
        return err;

    if ((cmd.response[0] & 0xff) != 0xaa)
        return UNUSABLE_ERR;
    else
        mmc->version = SD_VERSION_2;

    return 0;
}

int mmc_register(struct mmc *mmc)
{
    /* Setup the universal parts of the block interface just once */
    mmc->block_dev.if_type = IF_TYPE_MMC;
    mmc->block_dev.dev = cur_dev_num++;
    mmc->block_dev.removable = 1;
    mmc->block_dev.block_read = mmc_bread;
    mmc->block_dev.block_write = mmc_bwrite;
    mmc->block_dev.block_erase = mmc_berase;

#if defined(CONFIG_LG_CHG)
    mmc->block_dev.byte_read  = mmc_byteread;
    mmc->block_dev.byte_write = mmc_bytewrite;
	mmc->block_dev.berase_type = mmc_berase_type;
#endif	//#if defined(CONFIG_LG_CHG)

	mmc->block_dev.smart_report = mmc_get_smart_report;
	mmc->block_dev.wp_manage = mmc_wp_manage;

    if (!mmc->b_max)
        mmc->b_max = CONFIG_SYS_MMC_MAX_BLK_COUNT;

    INIT_LIST_HEAD (&mmc->link);

    list_add_tail (&mmc->link, &mmc_devices);

    return 0;
}

#ifdef CONFIG_PARTITIONS
block_dev_desc_t *mmc_get_dev(int dev)
{
    struct mmc *mmc = find_mmc_device(dev);
    if (!mmc)
        return NULL;

    mmc_init(mmc);
    return &mmc->block_dev;
}
#endif

int mmc_init(struct mmc *mmc)
{
    int err;

    if (mmc_getcd(mmc) == 0) {
        mmc->has_init = 0;
        printf("MMC: no card present\n");
        return NO_CARD_ERR;
    }

    if (mmc->has_init)
        return 0;

    err = mmc->init(mmc);

    if (err)
        return err;

    //mmc_set_bus_width(mmc, 1);
    //mmc_set_clock(mmc, 1);

#ifdef CONFIG_BSP_REALTEK // just decrease the initialized time

    /* Reset the Card */
    err = mmc_go_idle(mmc);

    if (err)
        return err;

    /* The internal partition reset to user partition(0) at every CMD0*/
    mmc->part_num = 0;

    err = mmc_send_op_cond(mmc);

    if (err) {
        printf("eMMC did not respond to voltage select!\n");
        printf("Try SD !\n");

        /* Test for SD version 2 */
        err = mmc_send_if_cond(mmc);

        /* Now try to get the SD card's operating condition */
        err = sd_send_op_cond(mmc);

    }

    if (err == TIMEOUT) {
        printf("Card did not respond to voltage select!\n");
        return UNUSABLE_ERR;
    }

#else
    /* run standard uboot emmc driver flow */

    /* Reset the Card */
    err = mmc_go_idle(mmc);

    if (err)
        return err;

    /* The internal partition reset to user partition(0) at every CMD0*/
    mmc->part_num = 0;

    /* Test for SD version 2 */
    err = mmc_send_if_cond(mmc);

    /* Now try to get the SD card's operating condition */
    err = sd_send_op_cond(mmc);

    /* If the command timed out, we check for an MMC card */
    if (err == TIMEOUT) {
        err = mmc_send_op_cond(mmc);

        if (err) {
            printf("Card did not respond to voltage select!\n");
            return UNUSABLE_ERR;
        }
    }
#endif

#ifdef CONFIG_BSP_REALTEK
    mmc_data_sync(mmc);     //sync RCA
#endif
    err = mmc_startup(mmc);
    if (err)
        mmc->has_init = 0;
    else
        mmc->has_init = 1;

#ifdef CONFIG_BSP_REALTEK
    mmc_data_sync(mmc); //sync high_capacity
#endif
    return err;
}

/*
 * CPU and board-specific MMC initializations.  Aliased function
 * signals caller to move on
 */
static int __def_mmc_init(bd_t *bis)
{
    return -1;
}

int cpu_mmc_init(bd_t *bis) __attribute__((weak, alias("__def_mmc_init")));
int board_mmc_init(bd_t *bis) __attribute__((weak, alias("__def_mmc_init")));

void print_mmc_devices(char separator)
{
    struct mmc *m;
    struct list_head *entry;

    list_for_each(entry, &mmc_devices) {
        m = list_entry(entry, struct mmc, link);

        printf("%s: %d", m->name, m->block_dev.dev);

        if (entry->next != &mmc_devices)
            printf("%c ", separator);
    }

    printf("\n");
}

int get_mmc_num(void)
{
    return cur_dev_num;
}

int mmc_initialize(bd_t *bis)
{
    INIT_LIST_HEAD (&mmc_devices);
    cur_dev_num = 0;

    if (board_mmc_init(bis) < 0)
        cpu_mmc_init(bis);

#ifndef CONFIG_LESS_CONSOLE
    print_mmc_devices(',');
#endif

    return 0;
}

