/*----------------------------------------------------------------------------
*
* The confidential and proprietary information contained in this file may
* only be used by a person authorised under and to the extent permitted
* by a subsisting licensing agreement from ARM Limited or its affiliates.
*
*        (C) COPYRIGHT 2008-2017 ARM Limited or its affiliates.
*             ALL RIGHTS RESERVED
*             
* This entire notice must be reproduced on all copies of this file
* and copies of this file may only be made by a person if such person is
* permitted to do so under the terms of a subsisting license agreement
* from ARM Limited or its affiliates.
*
* Release   : r0p0-00eac0
*-----------------------------------------------------------------------------*/

/* 

BRIEF DESCRIPTION, PURPOSE AND STRATEGY
=======================================

This test puts the GPU in its "normal" operating mode with no content running to allow for power measurements (norm) to be made.


DEBUG HINTS IN CASE OF FAILURE
===========================
a) Check the AXI bus connections as this is the bus that is used to communicate with the Job manager.
b) Check that the GPU's reset signal is not asserted.
c) Check that the GPU input clock is toggling as expected.
d) Check that the DFT signals are disabled.
e) Check that MBIST is disabled.

*/

#include "MaliFns.h"
#include "MaliDefines_DVALIN.h"


int RunMaliTest_pwr_normal_DVALIN_part0 (void);
void testDelay (void);

static int mali_step;

int RunMaliTest_pwr_normal_DVALIN (int init) {
  if (init) { mali_step = 0; };
  switch (mali_step) {
    case 0  : return RunMaliTest_pwr_normal_DVALIN_part0();
    default : return 0;
  };
};

/* Mali_test_reg
   Test the register.
   access: 0=RW, 1=RO
*/
int Mali_test_reg(int regnum, int read_mask, int write_mask, int reset_value, int access) {
  int value = Mali_RdReg(regnum);
  if( value != reset_value ) {
    printf("FAILURE: Wrong reset value. Addr: 0x%08x Value: 0x%08x Expected: 0x%08x\n",
           regnum, value, reset_value);
    return -1;
  }
  if( access == 0 ) {
    Mali_WrReg(regnum, (0xffffffff & write_mask));
    value = Mali_RdReg(regnum) & read_mask;
    if( value != (0xffffffff & write_mask & read_mask) ) {
      printf("FAILURE: Wrong value. Addr: 0x%08x Value: 0x%08x Expected: 0x%08x\n",
             regnum, value, (0xffffffff & write_mask & read_mask));
      return -2;
    }
    Mali_WrReg(regnum, (0x12345678 & write_mask));
    value = Mali_RdReg(regnum) & read_mask;
    if( value != (0x12345678 & write_mask & read_mask) ) {
      printf("FAILURE: Wrong value. Addr: 0x%08x Value: 0x%08x Expected: 0x%08x\n",
             regnum, value, (0x12345678 & write_mask & read_mask));
      return -2;
    }
  }
  printf("Register %08x: Success!\n", regnum);
  return 0;
}

int RunMaliTest_pwr_normal_DVALIN_part0 (void) {
  int res = 0;

  int num_cores, i, j, skip_core;
  volatile unsigned int core_bitmap;

  printf("RUNNING TEST: pwr_normal\n");
  printf("  Purpose: Measure power off & on consumption of shader core(s)\n");
  printf("  Will check register read/write and reset value\n");

  // Get current configuration to allow testing all registers
  core_bitmap   = Mali_RdReg(0x0100);

  i = core_bitmap;
  num_cores = 0;
  while (i) {
    num_cores++;
    i &= i-1;
  }

  res |= Mali_test_reg(0x000, 0xffff0000, 0xffff0000, GPU_ID_VALUE, 1); /* GPU_ID */
  res |= Mali_test_reg(0x100, 0xffffffff, 0xffffffff,  core_bitmap, 1); /* SHADER_PRESENT_LO */
  res |= Mali_test_reg(0x110, 0xffffffff, 0xffffffff, 0x00000001, 1); /* TILER_PRESENT_LO */
  res |= Mali_test_reg(0x120, 0xffffffff, 0xffffffff, 0x00000001, 1); /* L2_PRESENT_LO */
  res |= Mali_test_reg(0x140, 0xffffffff, 0xffffffff, 0x00000000, 1); /* SHADER_READY_LO */
  res |= Mali_test_reg(0x150, 0xffffffff, 0xffffffff, 0x00000000, 1); /* TILER_READY_LO */
  res |= Mali_test_reg(0x160, 0xffffffff, 0xffffffff, 0x00000000, 1); /* L2_READY_LO */

  printf("Reading SHADER_PWRON/OFF_LO - expecting OFF state\n");
  while( Mali_RdReg(0x200) ) {}
  res |= Mali_test_reg(0x140, 0xffffffff, 0xffffffff, 0, 1); /* SHADER_READY_LO */

  printf ("Adding a delay to allow stabilization for power measurement (OFF STATE)\n");
  testDelay();
  printf ("End of delay\n");

  for(i=0; core_bitmap!=0; i++) {
    
    if (core_bitmap & 0x1) {
        skip_core = 0;
    } else {
        skip_core = 1;
    }

    core_bitmap = core_bitmap >> 1;
    j = ((i + num_cores) % 4);

    if (skip_core) {
      printf("Skipping shader core #%d\n", i);
    } else {
      printf("Powering on shader core #%d\n", i);
      Mali_WrReg(0x180, (1<<i)); /* SHADER_PWRON_LO */

      printf("Reading SHADER_PWRON/OFF_LO of shader core #%d - expecting ON state\n", i);
      while( Mali_RdReg(0x200) ) {}
      res |= Mali_test_reg(0x140, 0xffffffff, 0xffffffff, (1<<i), 1); /* SHADER_READY_LO */

      printf ("Adding a delay to allow stabilization for power measurement (ON STATE) of shader core #%d\n", i);
      testDelay();
      printf ("End of delay\n");

      printf("Powering off shader core #%d\n", i);
      Mali_WrReg(0x1c0, (1<<i)); /* SHADER_PWROFF_LO */

      printf("Reading SHADER_PWRON/OFF_LO of shader core #%d - expecting OFF state\n", i);
      while( Mali_RdReg(0x200) ) {}
      res |= Mali_test_reg(0x140, 0xffffffff, 0xffffffff, 0, 1); /* SHADER_READY_LO */

      printf ("Adding a delay to allow stabilization for power measurement (OFF STATE) of shader core #%d\n", i);
      testDelay();
      printf ("End of delay\n");								     
    }
  }
  mali_step++;
  return res;
}

void testDelay (void) {
  int n, tmp;
#ifdef MALI_MINIMUM_DELAYS
  for (n = 0;n < (2);n++) {
#else  
  for (n = 0;n < 1024;n++) {
#endif  
    tmp = 2 * n;
    tmp = tmp - n;
  }
}
